#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic
This sub-module of geodezyx.operational contains functions to download
gnss data and products from distant IGS servers. 
it can be imported directly with:
from geodezyx import operational
The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License
Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
#### Import the logger
import logging
import multiprocessing as mp
import os

import geodezyx.operational.download_utils as dlutils
#### geodeZYX modules
from geodezyx import conv, utils

log = logging.getLogger('geodezyx')


##########  END IMPORT  ##########


# _____________ _   _ ________   __  _____                      _                 _
# |  __ \|_   _| \ | |  ____\ \ / / |  __ \                    | |               | |
# | |__) | | | |  \| | |__   \ v /  | |  | | _____      ___ __ | | ___   __ _  __| | ___ _ __
# |  _  /  | | | . ` |  __|   > <   | |  | |/ _ \ \ /\ / / '_ \| |/ _ \ / _` |/ _` |/ _ \ '__|
# | | \ \ _| |_| |\  | |____ / . \  | |__| | (_) \ v  v /| | | | | (_) | (_| | (_| |  __/ |
# |_|  \_\_____|_| \_|______/_/ \_\ |_____/ \___/ \_/\_/ |_| |_|_|\___/ \__,_|\__,_|\___|_|


############################################################################
######## RINEX DOWNLOADER
############################################################################


def igs_sopac_server_legacy(stat, date):
    # plante si trop de requete
    urlserver = "ftp://garner.ucsd.edu/pub/rinex/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), rnxname)
    return url


def igs_cddis_server_legacy(stat, date, user="", passwd=""):
    # a privilegier
    urlserver = "ftp://gdc.cddis.eosdis.nasa.gov/gps/data/daily/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(
        urlserver, str(date.year), conv.dt2doy(date), date.strftime("%y") + "d", rnxname
    )
    if not passwd:
        passwd = "sakic@ipgp.fr"
    return url, user, passwd


def igs_cddis_nav_server_legacy(stat, date):
    # a privilegier
    urlserver = "ftp://cddis.gsfc.nasa.gov/gps/data/daily/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date, "n.Z")
    url = os.path.join(
        urlserver, str(date.year), conv.dt2doy(date), date.strftime("%y") + "n", rnxname
    )
    return url


def nav_bkg_server(stat, date):
    urlserver = "ftp://igs-ftp.bkg.bund.de/IGS/BRDC/"
    # ftp://igs-ftp.bkg.bund.de/IGS/BRDC/2024/082/BRDC00WRD_S_20240820000_01D_MN.rnx.gz
    rnxname = "BRDC00WRD_S_" + conv.dt2str(date, "%Y%j") + "0000_01D_MN.rnx.gz"
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), rnxname)
    return url


def nav_rob_server_legacy(stat, date):
    urlserver = "ftp://epncb.oma.be/pub/obs/BRDC/"
    # ftp://epncb.oma.be/pub/obs/BRDC/2018/BRDC00GOP_R_20180010000_01D_MN.rnx.gz
    rnxname = "BRDC00GOP_R_" + conv.dt2str(date, "%Y%j") + "0000_01D_MN.rnx.gz"
    url = os.path.join(urlserver, str(date.year), rnxname)
    return url


def rgp_ign_smn_server_legacy(stat, date):
    urlserver = "ftp://rgpdata.ign.fr/pub/data/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), "data_30", rnxname)
    return url


def rgp_ign_mlv_server(stat, date):
    urlserver = "ftp://rgpdata.ensg.eu/pub/data/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), "data_30", rnxname)
    return url


def rgp_ign_smn_1_hz_server_legacy(stat, date):
    urlserver = "ftp://rgpdata.ign.fr/pub/data/"

    urls = []

    for h in range(24):
        date_session = date
        date_session = date_session.replace(hour=h)

        log.info("%s session %s", date_session, h)
        rnxname = conv.statname_dt2rinexname(
            stat.lower(), date_session, session_a_instead_of_daily_session=1
        )
        url = os.path.join(
            urlserver, str(date.year), conv.dt2doy(date), "data_1", rnxname
        )

        urls.append(url)

    return urls


def unavco_server_legacy(stat, date):
    urlserver = "ftp://data-out.unavco.org/pub/rinex"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, "obs", str(date.year), conv.dt2doy(date), rnxname)
    return url


def renag_server(stat, date):
    urlserver = "ftp://renag.unice.fr/data/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), rnxname)
    return url


def uwiseismic_server(stat, date, user="", passwd=""):
    urlserver = "ftp://www2.uwiseismic.com/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, "rinex", str(date.year), conv.dt2doy(date), rnxname)
    return url, user, passwd


def orpheon_server_legacy(stat, date, user="", passwd=""):
    urlserver = "ftp://renag.unice.fr/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), rnxname)
    return url, user, passwd


def ovsg_server_legacy(stat, date, user="", passwd=""):
    if dt.datetime(2009, 1, 1) <= date <= dt.datetime(2014, 2, 10):
        urlserver = (
            "http://webobs.ovsg.univ-ag.fr/rawdata/GPS-GPSDATA.backtemp_20140210/"
        )
    else:
        urlserver = "http://webobs.ovsg.univ-ag.fr/rawdata/GPS/GPSDATA/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), "rinex", rnxname)
    return url, user, passwd


def geoaus_server_legacy(stat, date):
    """Geosciences Australia
    ex : ftp://ftp.ga.gov.au/geodesy-outgoing/gnss/data/daily/2010/10063/"""
    urlserver = "ftp://ftp.ga.gov.au/geodesy-outgoing/gnss/data/daily/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(
        urlserver, str(date.year), date.strftime("%y") + conv.dt2doy(date), rnxname
    )
    return url


def sonel_server_legacy(stat, date):
    """ex : ftp://ftp.sonel.org/gps/data/2015/001/"""
    urlserver = "ftp://ftp.sonel.org/gps/data/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), rnxname)
    return url


def ens_fr_legacy(stat, date):
    urlserver = "ftp://gnss.ens.fr/pub/public/crl/GPS/rinex/"
    rnxname = conv.statname_dt2rinexname(stat.lower(), date)
    url = os.path.join(urlserver, str(date.year), conv.dt2doy(date), rnxname)
    return url


def effective_save_dir(parent_archive_dir, stat, date, archtype="stat"):
    """
    INTERNAL_FUNCTION

    archtype =
        stat
        stat/year
        stat/year/doy
        year/doy
        year/stat
        week/dow
        OR only '/' for a dirty saving in the parent folder
        ... etc ..."""
    if archtype == "/":
        return parent_archive_dir

    if len(archtype) > 0 and archtype[0] == "/":
        log.warn(
            "The archive type description starts with a /, remove it to avoid an error"
        )

    out_save_dir = parent_archive_dir
    fff = archtype.split("/")
    year = str(date.year)
    doy = conv.dt2doy(date)
    _, _ = year, doy  ## simply to remove the unused linter warning...
    week, dow = conv.dt2gpstime(date)
    for f in fff:
        out_save_dir = os.path.join(out_save_dir, eval(f))
    return out_save_dir


def multi_downloader_rinex(**kwargs):
    log.warn(
        "multi_downloader_rinex is a legacy alias for the newly renamed function download_gnss_rinex"
    )
    return download_gnss_rinex_legacy(**kwargs)


def download_gnss_rinex_legacy(
    statdico,
    archive_dir,
    startdate,
    enddate,
    archtype="stat",
    parallel_download=4,
    sorted_mode=False,
    user="",
    passwd="",
    filter_ftp_crawler=True,
    path_ftp_crawled_files_save=None,
    path_ftp_crawled_files_load=None,
    quiet_mode=False,
    final_archive_for_sup_check=None,
    force=False,
):
    """
    Download GNSS RINEX files from various archive centers.

    Parameters
    ----------
    statdico : dict
        Dictionary associating archive centers to list of stations.
        Format: statdico['archive_center'] = ['STA1', 'STA2', ...]
        
        Supported archive centers:
        - 'igs_cddis' or 'igs' : CDDIS center
        - 'igs_sopac' : SOPAC/UCSD/SIO center (not very reliable)
        - 'rgp' : IGN's RGP St Mandé center
        - 'rgp_mlv' : IGN's RGP Marne la Vallée center  
        - 'rgp_1Hz' : IGN's RGP, downloads all 24 hourly RINEX for the day
        - 'renag' : RENAG network
        - 'ovsg' : OVSG network
        - 'unavco' : UNAVCO center
        - 'sonel' : SONEL network
        - 'geoaus' : Geosciences Australia
        - 'ens_fr' : ENS France
        - 'nav' or 'brdc' : Navigation files from ROB server using GOP files
        - 'nav_rt' or 'brdc_rt' : Real-time navigation files from BKG server
    archive_dir : str
        Parent archive directory where files will be stored.
    startdate : datetime
        Start date of the wished period.
    enddate : datetime  
        End date of the wished period.
    archtype : str, optional
        String describing how the archive directory is structured.
        Examples: 'stat', 'stat/year', 'stat/year/doy', 'year/doy', 
        'year/stat', 'week/dow/stat', etc. Default is "stat".
    parallel_download : int, optional
        Number of parallel downloads. Default is 4.
    sorted_mode : bool, optional
        If False, uses the map multiprocess function so the download order 
        will be scrambled. If True, uses the apply multiprocess function so 
        the download order will be in chronological order. The scrambled 
        (False) is better because it doesn't create zombie processes. 
        Default is False.
    user : str, optional
        Username for locked servers. Default is "".
    passwd : str, optional
        Password for locked servers. Default is "".
    filter_ftp_crawler : bool, optional
        Use an improved FTP crawler to find which files actually exist
        to accelerate the download. If path_ftp_crawled_files_load is given,
        use this loaded list. Default is True.
    path_ftp_crawled_files_save : str, optional
        Path (directory + filename) to save a pickle containing the list of 
        existing RINEXs found on the server by the FTP crawler. Allows using 
        this list directly if facing a timeout during the download part.
        Default is None.
    path_ftp_crawled_files_load : str, optional
        Load and use the list of existing RINEXs found on the FTP server,
        generated by a previous run of the FTP crawler. Overrides an internal 
        call of crawl_ftp_files. Default is None.
    quiet_mode : bool, optional
        List the available RINEXs without downloading them. Useful only if 
        path_ftp_crawled_files_save is given. Default is False.
    final_archive_for_sup_check : str, optional
        The final archive path or a file containing the archived RINEXs in
        their final destination. Useful if the final archive is different 
        from archive_dir. Default is None.
    force : bool, optional
        Force the download even if the file already exists locally. 
        Default is False.

    Returns
    -------
    url_list : list of str
        List of URLs.
    savedir_list : list of str
        List of downloaded products paths.

    Examples
    --------
    >>> statdic = dict()
    >>> statdic['igs_cddis'] = ['ZIMM']
    >>> output_dir = '/home/USER/test_dl_rnx'
    >>> startdate = dt.datetime(2000, 1, 1)
    >>> enddate = dt.datetime(2000, 1, 31)
    >>> geodezyx.operational.download_gnss_rinex_legacy(statdic,
    """
    
    curdate = startdate

    pool = mp.Pool(processes=parallel_download)

    urllist = []
    savedirlist = []

    log.info("generating the list of potential RINEXs ...")

    while curdate <= enddate:
        for netwk, statlis in list(statdico.items()):

            if not utils.is_iterable(statlis):
                log.warning(
                    "%s given in the 'statdico' should be a list and it is not.",
                    statlis,
                )

            for stat in statlis:
                stat = stat.lower()
                mode1hz = False
                secure_ftp = False

                if netwk in ("igs_cddis", "igs"):
                    secure_ftp = True
                    url = igs_cddis_server_legacy(stat, curdate, user, passwd)
                elif netwk == "igs_sopac":
                    url = igs_sopac_server_legacy(stat, curdate)
                elif netwk == "rgp":
                    url = rgp_ign_smn_server_legacy(stat, curdate)
                elif netwk == "rgp_mlv":
                    url = rgp_ign_mlv_server(stat, curdate)
                elif netwk == "rgp_1Hz":
                    urls = rgp_ign_smn_1_hz_server_legacy(stat, curdate)
                    mode1hz = True
                elif netwk == "renag":
                    url = renag_server(stat, curdate)
                elif netwk == "orpheon":
                    url = orpheon_server_legacy(stat, curdate, user, passwd)
                elif netwk == "uwiseismic":
                    url = uwiseismic_server(stat, curdate, user, passwd)
                elif netwk == "ovsg":
                    url = ovsg_server_legacy(stat, curdate, user, passwd)
                elif netwk == "unavco":
                    url = unavco_server_legacy(stat, curdate)
                elif netwk == "sonel":
                    url = sonel_server_legacy(stat, curdate)
                elif netwk == "geoaus":
                    url = geoaus_server_legacy(stat, curdate)
                elif netwk in ("nav", "brdc"):
                    url = nav_rob_server_legacy(stat, curdate)
                elif netwk in ("nav_rt", "brdc_rt"):
                    url = nav_bkg_server(stat, curdate)
                elif netwk == "ens_fr":
                    url = ens_fr_legacy(stat, curdate)
                else:
                    log.warning("unkwn server dic in the dico, skip ...")
                    continue

                savedir = effective_save_dir(archive_dir, stat, curdate, archtype)

                if not mode1hz:
                    urllist.append(url)
                    savedirlist.append(savedir)
                else:
                    urllist = urllist + urls
                    savedirlist = savedirlist + [savedir] * len(urls)

        curdate = curdate + dt.timedelta(days=1)

    # savedirlist = [x for (y,x) in sorted(zip(urllist,savedirlist))]
    # urllist     = sorted(urllist)

    try:
        urllist, savedirlist = utils.sort_binom_list(urllist, savedirlist)
    except TypeError as err:
        log.error("unable to sort the URL and the save directory")
        log.error(
            "TIP: you maybe asked for servers with & without password in the same statdico"
        )
        raise err

    log.info(" ... done")
    log.info(str(len(urllist)) + " potential RINEXs")

    ### Use of the advanced FTP Crawler
    if filter_ftp_crawler:
        if path_ftp_crawled_files_load:  ## if the previous files are loaded
            urllist, savedirlist = utils.pickle_loader(path_ftp_crawled_files_load)
        else:  ## regular case
            urllist, savedirlist = dlutils.ftp_files_crawler_legacy(
                urllist, savedirlist, secure_ftp=secure_ftp
            )
            if path_ftp_crawled_files_save:
                savetup = (urllist, savedirlist)
                utils.pickle_saver(savetup, full_path=path_ftp_crawled_files_save)

    ##### Check if the rinex file already exists in the final archive
    if final_archive_for_sup_check and not force:
        Files_final_arch = utils.find_recursive(final_archive_for_sup_check, "*")
        Files_final_arch_basename = [os.path.basename(e) for e in Files_final_arch]

        urllist_new, savedirlist_new = [], []

        for u, sd in zip(urllist, savedirlist):
            if not os.path.basename(u) in Files_final_arch_basename:
                urllist_new.append(u)
                savedirlist_new.append(sd)

        urllist, savedirlist = urllist_new, savedirlist_new

    if not quiet_mode:
        if not secure_ftp:  ### all the servers except CDDIS
            if sorted_mode:
                _ = [
                    pool.apply_async(dlutils.downloader, args=(u, sd, force))
                    for u, sd in zip(urllist, savedirlist)
                ]
            else:
                forcelis = [force] * len(urllist)
                _ = pool.map(
                    dlutils.downloader_wrap, list(zip(urllist, savedirlist, forcelis))
                )
        else:  ## secure FTP i.e. CDDIS
            ftp_obj, _ = dlutils.ftp_objt_create(
                secure_ftp_inp=secure_ftp,
                host=url[0].split("/")[2],
                user=url[1],
                passwd=url[2],
            )

            for iurl, isavedir in zip(urllist, savedirlist):
                localpath, bool_dl = dlutils.ftp_downld_mono(ftp_obj, iurl[0], isavedir)

    localfiles_lis = []
    skiped_url = 0
    for url, savedir in zip(urllist, savedirlist):
        try:
            if type(url) is tuple:
                url0 = url[0]
            else:
                url0 = url
            localfile = os.path.join(savedir, os.path.basename(url0))
            if os.path.isfile(localfile):
                localfiles_lis.append(localfile)
        except Exception as e:
            # because of a weird error
            # i = p.rfind('/') + 1
            # AttributeError: 'tuple' object has no attribute 'rfind'
            skiped_url += 1
            log.warning(e)
            continue

    pool.close()
    if skiped_url > 0:
        log.debug(str(skiped_url) + " returned url skipped because of an Exception")
    return localfiles_lis, savedirlist


#    return zip(urllist,savedirlist)


def rnx_long2short_name(longname_filepath_in):
    """
    MUST BE IMPROVED
    """

    longname_basename = os.path.basename(longname_filepath_in)
    longname_dirname = os.path.dirname(longname_filepath_in)

    Longname_basename_splitted = longname_basename.split("_")

    datepart_str = Longname_basename_splitted[2]
    yyyy = datepart_str[:4]
    ddd = datepart_str[4:7]

    shortname_basename = longname_basename[:4].lower() + ddd + "0." + yyyy[2:] + "o"

    return os.path.join(longname_dirname, shortname_basename)


def multi_archiver_rinex(
    rinex_lis, parent_archive_dir, archtype="stat", move=True, force_mv_or_cp=False
):
    """
    Move or copy RINEX files to an archive directory structure.
    
    Takes a list of RINEX files and organizes them in the parent_archive_dir 
    according to the specified archive type structure.

    Parameters
    ----------
    rinex_lis : list of str
        List of RINEX file paths (typically generated by multi_finder_rinex).
    parent_archive_dir : str
        Parent directory where files will be archived.
    archtype : str, optional
        String describing how the archive directory is structured.
        Examples: 'stat', 'stat/year', 'stat/year/doy', 'year/doy', 
        'year/stat', 'week/dow/stat', etc. Default is "stat".
    move : bool, optional
        If True, move the files. If False, copy the files. Default is True.
    force_mv_or_cp : bool, optional
        If True, force move/copy even if the file already exists in the 
        destination. Default is False.

    Returns
    -------
    None

    Notes
    -----
    The function logs the number of files processed, skipped, and moved/copied.
    Directory structure is created automatically
    """

    mv_cnt = 0
    skip_cnt = 0
    log.info("RINEXs as input : %s", len(rinex_lis))

    if move:
        mv_fct = utils.move
    else:
        mv_fct = utils.copy2

    for rnx in rinex_lis:
        date = conv.rinexname2dt(rnx)
        rnxname = os.path.basename(rnx)
        stat = rnxname[0:4]
        savedir = effective_save_dir(parent_archive_dir, stat, date, archtype)
        utils.create_dir(savedir)

        if not force_mv_or_cp and os.path.isfile(os.path.join(savedir, rnxname)):
            skip_cnt += 1
            continue
        else:
            mv_fct(rnx, savedir)
            mv_cnt += 1

    log.info("RINEXs skiped :" + str(skip_cnt) + " (because already exist)")
    log.info("RINEXs moved  :" + str(mv_cnt))

    return None



# -*- coding: utf-8 -*-
"""
"""
import wget



def download_VMF(year=None,month=None,day=None): 
# =============================================================================
#  This subroutine downloads VMF files from ^1e VMF ^^rver. Here, only operational VMF1 data downloads. 
#  To download VMF1_FC, 'VMF_OP' should be changed to 'VMF1_FC'. 

#  File created by Zohreh Adavi zohreh.adavi@tuwien.ac.at, 2023-02-20 
# =============================================================================  
   LISTt=['00','06','12','18']
   for lstnum in range(0,len(LISTt)):
         try:
            filename='VMFG_'+str(year)+str(month)+str(day).zfill(2)+'.H'+LISTt[lstnum] 
            url="https://vmf.geo.tuwien.ac.at/trop_products/GRID/2.5x2/VMF1/VMF1_OP/"+str(year)+'/'+filename
            wget.download(url)
         except:
             pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  3 15:45:36 2024

@author: psakic
"""

import datetime as dt
import logging
import os

import pandas as pd
import xarray as xr

from geodezyx import operational, conv
from geodezyx.marine import obp

log = logging.getLogger('geodezyx')


def dac_get_aviso(date_srt_inp, date_end_inp, out_dir, user, passwd):
    """
    Downloads DAC (Doppler-derived Atmospheric Correction) files from the AVISO FTP server for a given date range.

    Parameters
    ----------
    date_srt_inp : datetime
        The start date of the range for which DAC files are to be downloaded.
    date_end_inp : datetime
        The end date of the range for which DAC files are to be downloaded.
    out_dir : str
        The local directory where the downloaded DAC files should be saved.
    user : str
        The username for the FTP server.
    passwd : str
        The password for the FTP server.

    Returns
    -------
    list
        A list of paths to the downloaded DAC files.

    """

    date_rng = pd.date_range(date_srt_inp, date_end_inp, freq="6H")

    ftpbase = "ftp://ftp-access.aviso.altimetry.fr/auxiliary/dac/dac_delayed_global/"
    p_stk = []
    for d in date_rng:
        jjul_cnes = conv.dt2jjul_cnes(d)
        dac_name = "dac_dif_" + str(jjul_cnes) + "_" + str(d.hour).zfill(2) + ".nc"

        p = ftpbase + str(d.year) + "/" + dac_name

        p_stk.append(p)

    operational.ftp_downld_front(p_stk, out_dir,
                                 user=user,
                                 passwd=passwd,
                                 secure_ftp=False)

    return


def dac_extract(dac_files_lis,
                mlon=45.46,
                mlat=-12.78,
                epoch_min=dt.datetime(1980, 1, 1),
                epoch_max=dt.datetime(2099, 1, 1)):
    """
    Extracts DAC (Doppler-derived Atmospheric Correction) values from a list
    of DAC files for a given location and date range.

    Parameters
    ----------
    dac_files_lis : list
        The list of DAC files from which the DAC values are to be extracted.
    mlon : float, optional
        The longitude of the location for which the DAC values are to be extracted. Default is 45.46.
    mlat : float, optional
        The latitude of the location for which the DAC values are to be extracted. Default is -12.78.
    epoch_min : datetime, optional
        The start date of the range for which the DAC values are to be extracted. Default is January 1, 1980.
    epoch_max : datetime, optional
        The end date of the range for which the DAC values are to be extracted. Default is January 1, 2099.

    Returns
    -------
    DataFrame
        A DataFrame with the extracted DAC values. The index of the DataFrame is the date and the column contains
        the DAC values.
    """

    if not dac_files_lis:
        log.error("No DAC files provided.")
        raise FileNotFoundError("No DAC files provided.")

    ds_stk = []
    dac_stk = []

    for f in dac_files_lis:
        log.debug("read DAC file: %s", f)
        ds = xr.open_dataset(f)
        bn = os.path.basename(f)
        # tname = conv.jjul_cnes2dt(bn[8:13]) + dt.timedelta(hours = int(bn[14:16]))
        dac_out = obp.interp_xy(ds, x=mlon, y=mlat).dac

        dac_val = float(dac_out.values)
        t = dac_out
        t = conv.str_date2dt(dac_out.date)

        ds_stk.append(ds)
        dac_stk.append((t, dac_val))

    df_dac_raw = pd.DataFrame(dac_stk)
    df_dac_raw.set_index(0, inplace=True)

    if epoch_max and epoch_min:
        df_dac = df_dac_raw.loc[epoch_min:epoch_max, 1]
    else:
        df_dac = df_dac_raw

    return df_dac



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.operational contains functions to send runs 
to the GFZ's cluster in a batch mode. 

it can be imported directly with:
from geodezyx import operational

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""


########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import getpass
#### Import the logger
import logging
import os
import re
import subprocess
import time

import numpy as np

#### geodeZYX modules
from geodezyx import utils

#### Import star style
# from geodezyx import *                   # Import the GeodeZYX modules
# from geodezyx.externlib import *         # Import the external modules
# from geodezyx.megalib.megalib import *   # Import the legacy modules names
log = logging.getLogger('geodezyx')

##########  END IMPORT  ##########
def cluster_GFZ_run(
    commands_list,
    bunch_on_off=True,
    bunch_job_nbr=10,
    bunch_wait_time=600,
    bj_check_on_off=True,
    bj_check_mini_nbr=2,
    bj_check_wait_time=120,
    bj_check_user="auto",
    add_cjob_cmd_prefix=True,
    wait_sleeping_before_launch=5,
):
    """
    Parameters
    ----------
    commands_list : list of str
        List of commands you want to run (one command per job).
    bunch_on_off : bool, optional
        If False, send all the jobs to the cluster at once.
        If True, just send <bunch_job_nbr> of them.
        See the options below.
        The default is True.
    bunch_job_nbr : int, optional
        number of jobs which will be sent to the cluser at once.
        The default is 10.
    bunch_wait_time : int, optional
        Minimal time between two bunch runs in sec.
        The default is 600.
    bj_check_on_off : bool, optional
        Do a check before a new bunch run, if previous jobs are still running.
        The default is True.
    bj_check_mini_nbr : int, optional
        The fuctions will wait <bj_check_wait_time> sec
        more if <bj_check_mini_nbr> or more jobs are running.
        The default is 2.
    bj_check_wait_time : int, optional
        wait time between two checks in sec. The default is 120.
    bj_check_user : str, optional
        username you want to check in the batchjobs.
        The default is "auto" i.e. your own username.
    add_cjob_cmd_prefix : bool, optional
        If, the input commands in command_list do not contain
        the cjob prefix command, it will be added automatically
        The default is True.
    wait_sleeping_before_launch : int, optional
        Waiting time between two successive runs.
        (To cancel the run if necessary)
        The default is 5.

    Returns
    -------
    None.

    """

    username = getpass.getuser()

    history_file_path = None

    i_bunch = 0

    if bj_check_user == "auto":
        bj_check_user = utils.get_username()

    log_path = "/home/" + bj_check_user + "/test_tmp.log"
    LOGobj = open(log_path, "w+")

    if not add_cjob_cmd_prefix:
        commands_list_opera = commands_list
    else:
        commands_list_opera = []
        for cmd in commands_list:
            cmd_ope = "cjob -c '" + cmd + "'"
            commands_list_opera.append(cmd_ope)

    log.info("****** JOBS THAT WILL BE LAUNCHED ******")
    log.info("Number of jobs : " + str(len(commands_list_opera)))
    log.info("****************************************")

    for ikommand, kommand in enumerate(commands_list_opera):

        ########## LOG/PRINT command
        log.info(kommand)
        LOGobj.write(kommand + "\n")

        ########## LOG/PRINT sleep
        info_sleep = (
            "script is sleeping for "
            + str(wait_sleeping_before_launch)
            + "sec (so you can cancel it) "
        )
        log.info(info_sleep)
        LOGobj.write(info_sleep + "\n")
        time.sleep(wait_sleeping_before_launch)

        ########## LOG/PRINT start
        info_start = "script starts @ " + str(dt.datetime.now())
        log.info(info_start)
        LOGobj.write(info_start + "\n")

        ########## RUN command here !!
        if True:
            p = subprocess.Popen(
                "",
                executable="/bin/csh",
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = p.communicate(kommand.encode())
        else:
            os.system(kommand)

        if history_file_path:
            with open(history_file_path, "a") as myfile:
                myfile.write(kommand + "\n")

        i_bunch += 1

        ########## Bunch On/Off : check if a bunch of job has been launched
        if bunch_on_off and np.mod(i_bunch, bunch_job_nbr) == 0:
            info_bunch = (
                "sleeping @ "
                + str(dt.datetime.now())
                + " for "
                + str(bunch_wait_time)
                + "s b.c. a bunch of "
                + str(bunch_job_nbr)
                + " jobs has been launched"
            )
            log.info(info_bunch)
            time.sleep(bunch_wait_time)
            LOGobj.write(info_bunch + "\n")

            ########## Bunch On/Off Check : check if the bunch is finished (experimental but on is better)
            ###### THIS PART MUST BE MERGE WITH THE SMALLER FCT BELLOW
            if log.info:
                log.info(
                    "All jobs should be finished now, let's see if there is some latecomers"
                )
                bj_check_tigger = False
            while not bj_check_tigger:
                bj_check_tigger = sleep_job_user(
                    minjob=bj_check_mini_nbr, bj_check_wait_time=bj_check_wait_time
                )

    return None


def number_job_user(bj_check_user=None, verbose=True):
    """
    Internal function for sleep_job_user
    """
    username = getpass.getuser()

    if not bj_check_user:
        bj_check_user = utils.get_username()

    bj_command = (
        "perl /dsk/igs2/soft_wrk/"
        + username
        + "/SOFT_EPOS8_BIN_TOOLS/SCRIPTS/e8_bjobs_local.pl"
    )

    bj_list = subprocess.check_output(bj_command, shell="/bin/csh")
    bj_list = bj_list.decode("utf-8")
    bj_list = bj_list.split("\n")

    bj_pattern_checked = bj_check_user + " *" + bj_check_user

    bj_list_checked = [bool(re.search(bj_pattern_checked, l)) for l in bj_list]
    bj_list_checked_sum = np.sum(bj_list_checked)

    if verbose:
        log.info(bj_list_checked_sum, "running jobs found for", bj_check_user)

    return bj_list_checked_sum, bj_list_checked, bj_pattern_checked


def sleep_job_user(bj_check_user=None, minjob=20, bj_check_wait_time=20):
    """
    Internal function for cluster_GFZ_run
    """
    if not bj_check_user:
        bj_check_user = utils.get_username()

    n_job, bj_list_checked, bj_pattern_checked = number_job_user(bj_check_user)

    if n_job >= minjob:
        check_tigger = False
        log.info(
            "sleeping @ "
            + str(dt.datetime.now())
            + " for "
            + str(bj_check_wait_time)
            + "s b.c."
            + str(n_job)
            + " jobs are runing"
        )
        time.sleep(bj_check_wait_time)
    else:
        check_tigger = True
        log.info("let's continue, no job matchs the pattern " + bj_pattern_checked)

    return check_tigger



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 13 15:11:12 2020

@author: psakic

This sub-module of geodezyx.conv deals with coordinate conversion.

it can be imported directly with:
from geodezyx import conv

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""


#### Import the logger
import logging

########## BEGIN IMPORT ##########
#### External modules
import numpy as np

#### geodeZYX modules
from geodezyx import utils
from geodezyx.conv import conv_rotation_matrices as rotmat

# import re
log = logging.getLogger("geodezyx")

##########  END IMPORT  ##########
#### Coordinates conversion

#   _____                    _ _             _               _____                              _
#  / ____|                  | (_)           | |             / ____|                            (_)
# | |     ___   ___  _ __ __| |_ _ __   __ _| |_ ___  ___  | |     ___  _ ____   _____ _ __ ___ _  ___  _ __
# | |    / _ \ / _ \| '__/ _` | | '_ \ / _` | __/ _ \/ __| | |    / _ \| '_ \ \ / / _ \ '__/ __| |/ _ \| '_ \
# | |___| (_) | (_) | | | (_| | | | | | (_| | ||  __/\__ \ | |___| (_) | | | \ v /  __/ |  \__ \ | (_) | | | |
#  \_____\___/ \___/|_|  \__,_|_|_| |_|\__,_|\__\___||___/  \_____\___/|_| |_|\_/ \___|_|  |___/_|\___/|_| |_|
#
#


def geo2xyz(lat, lon, h, angle="deg", e2=0.00669438003):
    """
    Coordinates conversion

    FLH Geographic => XYZ ECEF Cartesian

    Parameters
    ----------

    lat,lon,h : numpy.array of floats
        latitude (deg/rad), longitude (deg/rad), height (m)

    angle : string
        describe the angle input type : 'deg' or 'rad'

    e2 : floats
        ellipsoid parameters (WGS84 per default)


    Returns
    -------
    X,Y,Z : numpy.array of floats
        cartesian coordinates X,Y,Z in meters

    References
    ----------
    Based on PYACS of J.-M. Nocquet
    """

    if angle == "deg":
        lon = np.deg2rad(lon)
        lat = np.deg2rad(lat)

    # f=1.0 - sqrt(1-e2)

    wn = wnorm(lat)

    x = (wn + h) * np.cos(lat) * np.cos(lon)
    y = (wn + h) * np.cos(lat) * np.sin(lon)
    z = (wn * (1.0 - e2) + h) * np.sin(lat)

    return x, y, z


def xyz2geo(x, y, z, outdeg=True, a=6378137.0, e2=0.00669438003):
    """
    Coordinates conversion

    XYZ ECEF Cartesian => FLH Geographic

    Parameters
    ----------
    x,y,z : numpy.array of floats
        cartesian coordinates X,Y,Z in meters

    outdeg : bool
        if True, degrees output for the angle, else radian

    a, e2 : floats
        ellipsoid parameters (WGS84 per default)


    Returns
    -------
    latitude, longitude, height : numpy.array of floats
        latitude (deg/rad), longitude (deg/rad), height (m)

    References
    ----------
    Based on PYACS of J.-M. Nocquet
    """

    f = 1.0 - np.sqrt(1 - e2)

    tp = np.sqrt(x**2 + y**2)
    r = np.sqrt(tp**2 + z**2)

    tmu = np.arctan2((z / tp) * ((1.0 - f) + (e2 * a) / r), 1)
    rlbda = np.arctan2(y, x)

    s3 = (np.sin(tmu)) ** 3
    c3 = (np.cos(tmu)) ** 3
    t1 = z * (1 - f) + e2 * a * s3
    t2 = (1 - f) * (tp - e2 * a * c3)

    rphi = np.arctan2(t1, t2)

    rhe = tp * (np.cos(rphi)) + z * (np.sin(rphi))
    rhe = rhe - a * (np.sqrt(1 - e2 * (np.sin(rphi) ** 2)))

    if outdeg:
        rphi = np.rad2deg(rphi)
        rlbda = np.rad2deg(rlbda)

    return rphi, rlbda, rhe

def xyz2enu(x, y, z, x0, y0, z0):
    """
    Coordinates conversion

    XYZ ECEF Geocentric => ENU Topocentic

    Parameters
    ----------
    x,y,z : numpy.array of floats
        cartesian coordinates X,Y,Z in meters

    x0,y0,z0 : floats or numpy.array of floats
        coordinate of the wished topocentric origin point in the geocentric frame
        if they are iterable arrays (of the same size as X,Y,Z)
        a different x0,y0,z0 will be applied for each X,Y,Z element

    Returns
    -------
    e,n,u : numpy.array of floats
        East North Up Component (m) w.r.t. x0,y0,z0

    References
    ----------
    https://gssc.esa.int/navipedia/index.php/Transformations_between_ECEF_and_ENU_coordinates
    """

    f0, l0, h0 = xyz2geo(x0, y0, z0, outdeg=True)
    d_x = np.array(x) - x0
    d_y = np.array(y) - y0
    d_z = np.array(z) - z0
    e, n, u = xyz2enu_core(d_x, d_y, d_z, f0, l0)
    return e, n, u


def xyz2enu_around_fix_pos(x, y, z):
    """
    Computes the mean of the x,y,z and
    return the topocentric ENU position around its mean position

    Parameters
    ----------
    x,y,z : numpy.array of floats
        cartesian coordinates x,y,z in meters

    Returns
    -------
    E,N,U : numpy.array of floats
        East North Up Component (m) w.r.t. the mean position

    x0 , y0 , z0 : floats
        coordinates of the mean position in the geocentic frame
    """
    x0 = np.nanmean(x)
    y0 = np.nanmean(y)
    z0 = np.nanmean(z)
    return xyz2enu(x, y, z, x0, y0, z0), np.array([x0, y0, z0])


def enu2xyz(e, n, u, x0, y0, z0, velocity_mode=False):
    """
    Coordinates conversion

    ENU Topocentic => XYZ ECEF Geocentric

    Parameters
    ----------
    e,n,u : numpy.array of floats
        cartesian coordinates E,N,U in meters

    x0,y0,z0 : floats or numpy.array of floats
        coordinate of the topocentric origin point in the geocentric frame
        if they are iterable arrays (of the same size as E,N,U)
        a different x0,y0,z0 will be applied for each E,N,U element
    velocity_mode : bool
        For the Velocity mode, coordinates of the topocentric origin point
        are NOT added at the end of the conversion
        Default is False

    Returns
    -------
    x,y,z : numpy.array of floats
        ECEF X,Y,Z Component (m)

    References
    ----------
    https://gssc.esa.int/navipedia/index.php/Transformations_between_ECEF_and_ENU_coordinates
    """

    if utils.is_iterable(e):
        if not utils.is_iterable(x0):  ### The input x0,y0,z0 is a single point
            n = len(e)
            x0, y0, z0 = [x0] * n, [y0] * n, [z0] * n
        else:  ### The input x0,y0,z0 is a vector = different x0,y0,z0 for each point
            x0, y0, z0 = x0, y0, z0

        xlist, ylist, zlist = [], [], []

        for e, n, u, x0use, y0use, z0use in zip(e, n, u, x0, y0, z0):
            x, y, z = enu2xyz(e, n, u, x0use, y0use, z0use, velocity_mode=velocity_mode)
            xlist.append(x)
            ylist.append(y)
            zlist.append(z)
        return np.array(xlist), np.array(ylist), np.array(zlist)

    else:
        fr, lr, hr = xyz2geo(x0, y0, z0)
        f0 = np.deg2rad(fr)
        l0 = np.deg2rad(lr)

        r = np.array(
            [
                [-np.sin(l0), np.cos(l0), 0],
                [-np.sin(f0) * np.cos(l0), -np.sin(f0) * np.sin(l0), np.cos(f0)],
                [np.cos(f0) * np.cos(l0), np.cos(f0) * np.sin(l0), np.sin(f0)],
            ]
        )

        #r3 = r.T
        r3 = np.linalg.inv(r)

        enu = np.vstack((e, n, u))

        xyz = np.dot(r3, enu)

        if velocity_mode:
            x = float(xyz[0])
            y = float(xyz[1])
            z = float(xyz[2])
        else:
            x = float(xyz[0]) + x0
            y = float(xyz[1]) + y0
            z = float(xyz[2]) + z0

        return x, y, z



def xyz2azi_ele(x, y, z, x0, y0, z0, outdeg=False):
    """
    Convert ECEF Cartesian coordinates to azimuth and elevation angles.

    Parameters
    ----------
    x, y, z : numpy.array of floats
        ECEF X, Y, Z coordinates in meters.
    x0, y0, z0 : floats or numpy.array of floats
        Reference point coordinates in ECEF frame.

    Returns
    -------
    azi : numpy.array of floats
        Azimuth angle in radians.
    ele : numpy.array of floats
        Elevation angle in radians.
    d : numpy.array of floats
        Distance from the reference point to the point (x, y, z) in meters.
    outdeg : bool, optional
        If True, output angles in degrees. If False, in radians. Default is False.
    """
    e, n, u = xyz2enu(x, y, z, x0, y0, z0)
    d = np.sqrt(e**2 + n**2 + u**2)
    azi = np.arctan2(e, n)
    ele = np.arcsin(u/d)
    #ele2 = np.arctan2(u, np.sqrt(e**2 + n**2))
    if outdeg:
        azi = np.rad2deg(azi)
        ele = np.rad2deg(ele)
    return azi, ele, d






 # __      __       _              __          __
 # \ \    / /      | |             \ \        / /
 #  \ \  / /__  ___| |_ ___  _ __   \ \  /\  / / __ __ _ _ __  _ __   ___ _ __ ___
 #   \ \/ / _ \/ __| __/ _ \| '__|   \ \/  \/ / '__/ _` | '_ \| '_ \ / _ \ '__/ __|
 #    \  /  __/ (__| || (_) | |       \  /\  /| | | (_| | |_) | |_) |  __/ |  \__ \
 #     \/ \___|\___|\__\___/|_|        \/  \/ |_|  \__,_| .__/| .__/ \___|_|  |___/
 #                                                      | |   | |
 #                                                      |_|   |_|


def geo2xyz_vector(llh, angle="deg", a=6378137.0, e2=0.00669438003):
    """
    Convert an array of geographic coordinates (latitude, longitude, height) to ECEF Cartesian coordinates.

    Parameters
    ----------
    llh : array-like
        Array of shape (N, 3) or (3, N) containing latitude, longitude, and height.
    angle : str, optional
        Specifies if input angles are in degrees ('deg') or radians ('rad'). Default is 'deg'.
    a : float, optional
        Semi-major axis of the ellipsoid in meters. Default is 6378137.0 (WGS84).
    e2 : float, optional
        Square of the first eccentricity of the ellipsoid. Default is 0.00669438003 (WGS84).

    Returns
    -------
    xyz : numpy.ndarray
        Array of shape (N, 3) with ECEF X, Y, Z coordinates.
    """
    #### if Nx3 array => 3xN array
    llh = utils.transpose_vector_array(llh)

    x, y, z = geo2xyz(llh[0], llh[1], llh[2], angle=angle, a=a, e2=e2)
    xyz = np.column_stack((x, y, z))

    return xyz


def xyz2enu_vector(xyz, xyz0):
    """
    Convert an array of ECEF Cartesian coordinates to ENU topocentric coordinates.

    Parameters
    ----------
    xyz : array-like
        Array of shape (N, 3) or (3, N) with ECEF X, Y, Z coordinates.
    xyz0 : array-like
        Reference point as array-like of length 3 (X0, Y0, Z0).

    Returns
    -------
    enu : numpy.ndarray
        Array of shape (N, 3) with East, North, Up coordinates.
    """
    xyz = utils.transpose_vector_array(xyz)

    e, n, u = xyz2enu(xyz[0], xyz[1], xyz[2], xyz0[0], xyz0[1], xyz0[2])
    enu = np.column_stack((e, n, u))

    return enu


def xyz2geo_vector(xyz, outdeg=True, a=6378137.0, e2=0.00669438003):
    """
    Convert an array of ECEF Cartesian coordinates to geographic coordinates (latitude, longitude, height).

    Parameters
    ----------
    xyz : array-like
        Array of shape (N, 3) or (3, N) with ECEF X, Y, Z coordinates.
    outdeg : bool, optional
        If True, output angles are in degrees. If False, in radians. Default is True.
    a : float, optional
        Semi-major axis of the ellipsoid in meters. Default is 6378137.0 (WGS84).
    e2 : float, optional
        Square of the first eccentricity of the ellipsoid. Default is 0.00669438003 (WGS84).

    Returns
    -------
    flh : numpy.ndarray
        Array of shape (N, 3) with latitude, longitude, and height.
    """
    xyz = utils.transpose_vector_array(xyz)

    f, l, h = xyz2geo(xyz[0], xyz[1], xyz[2], outdeg=outdeg, a=a, e2=e2)

    flh = np.column_stack((f, l, h))

    return flh


def enu2xyz_vector(enu, xyz_ref):
    """
    Convert an array of ENU topocentric coordinates to ECEF Cartesian coordinates.

    Parameters
    ----------
    enu : array-like
        Array of shape (N, 3) or (3, N) with East, North, Up coordinates.
    xyz_ref : array-like
        Reference point as array-like of length 3 (X0, Y0, Z0).

    Returns
    -------
    xyz : numpy.ndarray
        Array of shape (N, 3) with ECEF X, Y, Z coordinates.
    """
    enu = utils.transpose_vector_array(enu)

    x, y, z = enu2xyz(enu[0], enu[1], enu[2], xyz_ref[0], xyz_ref[1], xyz_ref[2])

    xyz = np.column_stack((x, y, z))

    return xyz

#      _                             __      _      _      _                                                _
#     (_)                           / /     | |    | |    | |                                              (_)
#  ___ _  __ _ _ __ ___   __ _     / /   ___| |_ __| |  __| | _____   __   ___ ___  _ ____   _____ _ __ ___ _  ___  _ __
# / __| |/ _` | '_ ` _ \ / _` |   / /   / __| __/ _` | / _` |/ _ \ \ / /  / __/ _ \| '_ \ \ / / _ \ '__/ __| |/ _ \| '_ \
# \__ \ | (_| | | | | | | (_| |  / /    \__ \ || (_| || (_| |  __/\ v /  | (_| (_) | | | \ v /  __/ |  \__ \ | (_) | | | |
# |___/_|\__, |_| |_| |_|\__,_| /_/     |___/\__\__,_(_)__,_|\___| \_(_)  \___\___/|_| |_|\_/ \___|_|  |___/_|\___/|_| |_|
#         __/ |
#        |___/

def sigma_xyz2enu(x, y, z, s_x, s_y, s_z, s_xy=0, s_yz=0, s_xz=0):
    """
    Convert standard deviation
    Cartesian ECEF XYZ => Cartesian Topocentric ENU

    Note
    ----
    Inputs values are now assumed correlated (241105)

    References
    ----------
    Linear Algebra, Geodesy, and GPS p332
    https://stackoverflow.com/questions/51162460/converting-ecef-xyz-covariance-matrix-to-enu-covariance-matrix
    https://gssc.esa.int/navipedia/index.php/Transformations_between_ECEF_and_ENU_coordinates
    """

    sigma_xyz = np.array(
        [[s_x**2, s_xy, s_xz], [s_xy, s_y**2, s_yz], [s_xz, s_yz, s_z**2]]
    )

    f, l, h = xyz2geo(x, y, z)

    # old and bad rotation matrix (bofore 20241104)
    # C = rotmat.C_ecef2enu(F,L,angtype='deg')
    # new and good rotation matrix (after 20241104)
    c = rotmat.c_ecef2enu_sigma(f, l, angtype="deg")

    sigma_enu = np.dot(np.dot(c, sigma_xyz), c.T)
    sigma_enu2 = np.dot(c, np.dot(sigma_xyz, c.T))

    # lon = np.deg2rad(L)
    # lat = np.deg2rad(F)

    s_e = np.sqrt(sigma_enu[0, 0])
    s_n = np.sqrt(sigma_enu[1, 1])
    s_u = np.sqrt(sigma_enu[2, 2])

    return s_e, s_n, s_u


def sigma_geo2xyz(lat, lon, h, s_f, s_l, s_h, ang="deg"):
    """
    Convert standard deviation
    Geographic FLH => Cartesian ECEF XYZ

    Warning
    -------
    Inputs values are assumed as uncorrelated, which is not accurate
    Must be improved

    References
    ----------
    Linear Algebra, Geodesy, and GPS p332
    """

    log.warning("Inputs values are assumed as uncorrelated, which is not accurate")
    log.warning("Prefer sigma_xyz2enu")

    if ang == "deg":
        lat = np.deg2rad(lat)
        lon = np.deg2rad(lon)
        s_f = np.deg2rad(s_f)
        s_l = np.deg2rad(s_l)
    x, y, z = geo2xyz(lat, lon, h)
    x2, y2, z2 = geo2xyz(lat + s_f, lon + s_l, h + s_h)

    return np.abs(x - x2), np.abs(y - y2), np.abs(z - z2)


def sigma_geo2enu(lat, lon, h, s_f, s_l, s_h, ang="deg"):
    """
    Convert standard deviation
    Geographic FLH => Cartesian Topocentric ENU

    Warning
    -------
    Inputs values are assumed as uncorrelated, which is not accurate
    Must be improved

    References
    ----------
    Linear Algebra, Geodesy, and GPS p332
    """
    # conversion batarde du sigma FLH => sigma ENU
    # Par conversion des angles en distance

    log.warning("Inputs values are assumed as uncorrelated, which is not accurate")
    log.warning("Prefer sigma_xyz2enu")

    if ang == "deg":
        lat = np.deg2rad(lat)
        lon = np.deg2rad(lon)
        s_f = np.deg2rad(s_f)
        s_l = np.deg2rad(s_l)

    e2 = 0.00669438003
    # f=1.0 - np.sqrt(1-e2)
    a = 6378137.0

    # Je présume que Rpsi est le rayon du cercle tangent à l'ellipsoide à une
    # lattitude donnée (comm du 150710)
    psi = np.arctan((1 - e2) * np.tan(lat))
    r_psi = a * np.sqrt(1 - e2) / np.sqrt(1 - e2 * np.cos(psi) ** 2)

    # r est le rayon d'un petit cercle (un parallèle)
    r = np.cos(psi) * (r_psi + h)

    # On pourrait simplifier par 2pi mais autant avoir toute la démarche
    s_e = (np.pi * 2 * r * s_l) / (2 * np.pi)
    s_n = (np.pi * 2 * r_psi * s_f) / (2 * np.pi)
    s_u = s_h
    return s_e, s_n, s_u


def sigma_enu2geo(lat, lon, h, s_e, s_n, s_u, ang="deg", a=6378137.0, e2=0.00669438003):
    """
    Convert standard deviation
    Cartesian Topocentric ENU => Geographic FLH

    Warning
    -------
    Inputs values are assumed as uncorrelated, which is not accurate
    Must be improved

    References
    ----------
    Linear Algebra, Geodesy, and GPS p332
    """

    log.warning("Inputs values are assumed as uncorrelated, which is not accurate")
    log.warning("Prefer sigma_xyz2enu")

    # conversion batarde du sigma ENU => sigma FLH
    # Par conversion des angles en distance

    # f=1.0 - np.sqrt(1-e2)

    # Je présume que Rpsi est le rayon du cercle tangent à l'ellipsoide à une
    # lattitude donnée (comm du 150710)
    psi = np.arctan((1 - e2) * np.tan(lat))
    r_psi = a * np.sqrt(1 - e2) / np.sqrt(1 - e2 * np.cos(psi) ** 2)

    # r est le rayon d'un petit cercle (un parallèle)
    r = np.cos(psi) * (r_psi + h)

    # On pourait simplifier par 2pi mais autant avoir toute la démarche
    s_l = s_e * (2 * np.pi) / (np.pi * 2 * r)
    s_f = s_n * (2 * np.pi) / (np.pi * 2 * r_psi)
    s_h = s_u

    if ang == "deg":
        s_l = np.rad2deg(s_l)
        s_f = np.rad2deg(s_f)
    return s_f, s_l, s_h


def eci2rtn_or_rpy(p, v, c, out_rpy=False, rpy_theo_mode=False):
    """
    convert ECI coordinates in RTN (RIC) or RPY (Roll Pitch Yaw)

    Parameters
    ----------
    p : numpy.array
        3D vector, position of the ref object in ECI frame
    v : numpy.array
        3D vector, velocity of the ref object in ECI frame
    c : numpy.array
        3D Vector, coordinates in ECF frame that will be transformed
    out_rpy : bool
        if True output in RPY frame, RTN instead
    rpy_theo_mode : bool
        use the theoretical matrix composition, but wrong
        ans. empirically, only for debug !!

    Returns
    -------
    c_out : conversion of p in RTN or RPY ref. frame

    References
    ----------
    "Coordinate Systems", ASEN 3200 1/24/06 George H. Born
    """

    c_eci2rtn_mat = rotmat.c_eci2rtn(p, v)

    if not out_rpy:
        trans_mat = c_eci2rtn_mat
    else:

        if not rpy_theo_mode:
            # Pour de très obscures raisons la composition est inversée
            # par rapport à l'ordre standard ... (241017)
            trans_mat = np.dot(rotmat.c_rtn2rpy().T, c_eci2rtn_mat.T)
        else:
            log.warning(
                "using the theoretical mode for RPY conversion, UNSTABLE & WRONG !"
            )
            trans_mat = np.dot(rotmat.c_rtn2rpy(), c_eci2rtn_mat)

        # Mais reste compatible avec Wikipedia
        # https://en.wikipedia.org/wiki/Permutation_matrix#Permutation_of_rows_and_columns

    c_out = trans_mat.dot(c)

    # EXEMPLE POUR DEBUG
    #     AAAAAAAAAAAAA yawerr,aux,nomi
    #       -1.606542300481316
    #       -1.606542300481316
    # CCCCCCCCCCCC sunors orientation du soleil
    #   2.379858191647978E-002   6.654859888132475E-001  -7.460308480163662E-001
    # DDDDDDDDDDDDD Inputs : xsat , vsat , xsun
    #   -15779.437215560150435   -23508.964967410494864     8644.453520237560952
    #        1.296220384065493       -1.917314252639830       -2.847293074880089
    #  -1.342404641197934E+008  -5.934080206107079E+007  -2.572342068927875E+007
    # EEEEEEEEEEEE Rcrs2ors
    #   3.532662370257567E-001  -7.688096793507927E-001   5.330195492501879E-001
    #  -5.225364448455557E-001   3.104619621186746E-001   7.941181766553120E-001
    #  -7.759888076421138E-001  -5.590572803191509E-001  -2.920042493231377E-001
    #    Rcrs2ors = np.array([[ 0.35330134, -0.76880968,  0.53301955],
    #                         [-0.52248415,  0.31046196,  0.79411818],
    #                         [-0.77600804, -0.55905728, -0.29200425]])

    #    Psat = np.array([   -15779.437215560150435 ,  -23508.964967410494864  ,   8644.453520237560952])
    #    Vsat = np.array([        1.296220384065493  ,     -1.917314252639830  ,     -2.847293074880089])
    #    Psun = np.array([  -1.342404641197934E+008 , -5.934080206107079E+007 , -2.572342068927875E+007])

    return c_out


def eci2rtn(p,v,c):
    """
    legacy wrapper of eci2rtn_or_rpy
    """
    return eci2rtn_or_rpy(p,v,c, out_rpy=False)


def ecef2eci(xyz, utc_times):
    """
    Convert ECEF (Earth Centered Earth Fixed) positions to ECI (Earth Centered Inertial)
    positions

    Parameters
    ----------
    xyz : numpy.array of floats
        XYZ are cartesian positions in ECEF. Should have shape (N,3)

    utc_times : numpy.array of floats
        UTC_times are UTC timestamps, as datetime objects. Sould have shape (N)

    Returns
    -------
    eci : numpy.array of floats
        Earth Centered Inertial coordinates. will have shape (N,3)

    Note
    ----
    Requires pyorbital module

     [X]    [C -S 0][X]
     [Y]  = [S  C 0][Y]
     [Z]eci [0  0 1][Z]ecf

     C and S are cos() and sin() of gmst (Greenwich Meridian Sideral Time)

    References
    ----------
    http://ccar.colorado.edu/ASEN5070/handouts/coordsys.doc
    Inspired from satellite-js (https://github.com/shashwatak/satellite-js)
    """

    # XYZ and utc_time must have the same shape
    # if not xyz.shape[:-1] == utc_times.shape:
    #    raise ValueError("shape mismatch for XYZ and utc_times (got {} and {})".format(xyz.shape[:-1],utc_times.shape))

    #    gmst = -1 * astronomy.gmst(utc_times) # EDIT 180430 : Why this -1 ??? removed because wrong ! ...
    from pyorbital import astronomy

    gmst = 1 * astronomy.gmst(utc_times)

    eci = xyz.copy()
    eci[:, 0] = xyz[:, 0] * np.cos(gmst) - xyz[:, 1] * np.sin(gmst)
    eci[:, 1] = xyz[:, 0] * np.sin(gmst) + xyz[:, 1] * np.cos(gmst)
    return eci


def eci2ecef(xyz, utc_times):
    """
    Convert ECI (Earth Centered Inertial) positions to ECEF (Earth Centered Earth Fixed)
    positions

    Parameters
    ----------
    xyz : numpy.array of floats
        XYZ are cartesian positions in Earth Centered Inertial. Should have shape (N,3)

    utc_times : numpy.array of floats
        UTC_times are UTC timestamps, as datetime objects. Sould have shape (N)

    Returns
    -------
    ecef : numpy.array of floats
        Earth Centered Earth Fixed coordinates. will have shape (N,3)

    Note
    ----
    Requires pyorbital module


     [X]          ([C -S 0])[X]
     [Y]     = inv([S  C 0])[Y]
     [Z]ecef      ([0  0 1])[Z]eci


    Empirically:
     [X]       [ C  S  0][X]
     [Y]     = [-S  C  0][Y]
     [Z]ecef   [ 0  0  1][Z]eci



     C and S are cos() and sin() of gmst (Greenwich Meridian Sideral Time)

    References
    ----------
    http://ccar.colorado.edu/ASEN5070/handouts/coordsys.doc
    Inspired from satellite-js (https://github.com/shashwatak/satellite-js)


    Note
    ----
    Quick mode of the reverse fct, can be improved
    """
    # XYZ and utc_time must have the same shape
    # if not xyz.shape[:-1] == utc_times.shape:
    #    raise ValueError("shape mismatch for XYZ and utc_times (got {} and {})".format(xyz.shape[:-1],utc_times.shape))

    #    gmst = -1 * astronomy.gmst(utc_times)
    # EDIT 180430 : Why this -1 ??? removed because wrong ! ...
    from pyorbital import astronomy

    gmst = 1 * astronomy.gmst(utc_times)

    ecef = xyz.copy()
    ecef[:, 0] = +xyz[:, 0] * np.cos(gmst) + xyz[:, 1] * np.sin(gmst)
    ecef[:, 1] = -xyz[:, 0] * np.sin(gmst) + xyz[:, 1] * np.cos(gmst)

    return ecef


#### Core & Legacy


 #   _____                          _
 #  / ____|                 ___    | |
 # | |     ___  _ __ ___   ( _ )   | |     ___  __ _  __ _  ___ _   _
 # | |    / _ \| '__/ _ \  / _ \/\ | |    / _ \/ _` |/ _` |/ __| | | |
 # | |___| (_) | | |  __/ | (_>  < | |___|  __/ (_| | (_| | (__| |_| |
 #  \_____\___/|_|  \___|  \___/\/ |______\___|\__, |\__,_|\___|\__, |
 #                                              __/ |            __/ |
 #                                             |___/            |___/

def vector_separator(abc):
    """
    Split a Nx3 Array/DataFrame in three separated 1-component vectors
    To simplify the usage of the conversion functions
    (which take single component vectors as input)

    Parameters
    ----------
    abc : Array or DataFrame
        Nx3 XYZ, ENU... array.

    Returns
    -------
    a : Array
        1st component.
    b : Array
        2nd component.
    c : Array
        3rd component.

    """
    abc = np.array(abc)
    return abc[:, 0], abc[:, 1], abc[:, 2]


def wnorm(phi, a=6378137.0, e2=0.00669438003):
    """
    Compute the Ellipsoid "Grande Normale"

    References
    ----------
    ALG0021 in NTG_71 (IGN Lambert Projection Tech document)
    Based on PYACS of J.-M. Nocquet
    """
    from numpy import sqrt, sin

    # e=sqrt(e2)
    wd = sqrt(1 - e2 * sin(phi) ** 2)
    result = a / wd
    return result


def normal_vector(phi, llambda, angle="deg", normalized=True):
    """
    Compute the Ellipsoid "Normale"

    References
    ----------
    p. Bosser (2012), Géométrie de l'Ellipsoïde, p27
    """

    if angle == "deg":
        llambda = np.deg2rad(llambda)
        phi = np.deg2rad(phi)

    a = np.cos(llambda) * np.cos(phi)
    b = np.sin(llambda) * np.cos(phi)
    c = np.sin(phi)

    n = np.array([a, b, c])

    if normalized:
        n = n / np.linalg.norm(n)

    return n

def xyz2enu_core(d_x, d_y, d_z, lat0, lon0):
    """
    Coordinates conversion

    XYZ ECEF Geocentric => ENU Topocentic

    **use xyz2enu in priority**

    (xyz2enu_core is a core function for xyz2enu)

    dXYZ = XYZrover - XYZref

    Parameters
    ----------
    d_x,d_y,d_z : floats or numpy.array of floats
        cartesian coordinates difference between the considered point(s)
        and the reference point

    lat0,lon0 : floats or numpy.array of floats
        if they are iterable arrays (of the same size as d_x,dY,dZ)
        a different x0,y0,z0 will be applied for each d_x,dY,dZ element

    Returns
    -------
    E,N,U : numpy.array of floats
        East North Up Component (m) w.r.t. x0,y0,z0

    References
    ----------
    https://gssc.esa.int/navipedia/index.php/Transformations_between_ECEF_and_ENU_coordinates

    Note
    ----
    This recursive fuction should be improved
    """

    ## Case one ref point per dXYZ
    if utils.is_iterable(lat0):
        e, n, u = [], [], []
        for dx_m, dy_m, dz_m, lat0_m, lon0_m in zip(d_x, d_y, d_z, lat0, lon0):
            e_m, n_m, u_m = xyz2enu_core(dx_m, dy_m, dz_m, lat0_m, lon0_m)
            e.append(e_m)
            n.append(n_m)
            u.append(u_m)

        return np.squeeze(np.array(e)), np.squeeze(np.array(n)), np.squeeze(np.array(u))

    # case onle ref point for all dXYZ
    else:
        f0 = np.deg2rad(lat0)
        l0 = np.deg2rad(lon0)

        r = np.array(
            [
                [-np.sin(l0), np.cos(l0), 0],
                [-np.sin(f0) * np.cos(l0), -np.sin(f0) * np.sin(l0), np.cos(f0)],
                [np.cos(f0) * np.cos(l0), np.cos(f0) * np.sin(l0), np.sin(f0)],
            ]
        )

        enu = np.dot(r, np.vstack((d_x, d_y, d_z)))

        e = enu[0, :]
        n = enu[1, :]
        u = enu[2, :]

        return e, n, u

def enu2xyz_legacy(e, n, u, x0, y0, z0):
    """
    KEPT FOR LEGACY REASONS, use enu2xyz

    diffère de enu2xyz pour d'obscure raisons, à investiguer !!!
    est laissé pour des scripts de conversion de GINS (170119)

    this fct compute the dXYZ and not the final XYZ

    """

    fr, lr, hr = xyz2geo(x0, y0, z0)
    f0 = np.deg2rad(fr)
    l0 = np.deg2rad(lr)

    r = np.array(
        [
            [-np.sin(l0), np.cos(l0), 0],
            [-np.sin(f0) * np.cos(l0), -np.sin(f0) * np.sin(l0), np.cos(f0)],
            [np.cos(f0) * np.cos(l0), np.cos(f0) * np.sin(l0), np.sin(f0)],
        ]
    )

    r3 = r.T

    enu = np.vstack((e, n, u))

    xyz = np.dot(r3, enu)  # + np.vstack(( x0, y0, z0))

    d_x = float(xyz[0])
    d_y = float(xyz[1])
    d_z = float(xyz[2])

    return d_x, d_y, d_z

############ deprecated aliases ############
import warnings

def deprec_warn(old, new):
    """
    Helper function to issue deprecation warnings
    """
    warnings.warn(
        f"{old} is deprecated and will be removed in future versions. Use {new} instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return None


def ECI2ECEF(*args, **kwargs):
    """
    Deprecated alias for eci2ecef
    """
    deprec_warn("ECI2ECEF", "eci2ecef")
    return eci2ecef(*args, **kwargs)


def ECEF2ECI(*args, **kwargs):
    """
    Deprecated alias for ecef2eci
    """
    deprec_warn("ECEF2ECI", "ecef2eci")
    return ecef2eci(*args, **kwargs)


def XYZ2ENU_2(*args, **kwargs):
    """
    Deprecated alias for xyz2enu
    """
    deprec_warn("XYZ2ENU_2", "xyz2enu")
    return xyz2enu(*args, **kwargs)


def ENU2XYZ(*args, **kwargs):
    """
    Deprecated alias for enu2xyz
    """
    deprec_warn("ENU2XYZ", "enu2xyz")
    return enu2xyz(*args, **kwargs)


def GEO2XYZ(*args, **kwargs):
    """
    Deprecated alias for geo2xyz
    """
    deprec_warn("GEO2XYZ", "geo2xyz")
    return geo2xyz(*args, **kwargs)


def XYZ2GEO(*args, **kwargs):
    """
    Deprecated alias for xyz2geo
    """
    deprec_warn("XYZ2GEO", "xyz2geo")
    return xyz2geo(*args, **kwargs)


def sXYZ2sENU(*args, **kwargs):
    deprec_warn("sXYZ2sENU", "sigma_xyz2enu")
    """
    Deprecated alias for sigma_xyz2enu
    """
    return sigma_xyz2enu(*args, **kwargs)


def sENU2sFLH(*args, **kwargs):
    """
    Deprecated alias for sigma_enu2geo
    """
    deprec_warn("sENU2sFLH", "sigma_enu2geo")
    return sigma_enu2geo(*args, **kwargs)

def sFLH2sENU(*args, **kwargs):
    """
    Deprecated alias for sigma_geo2enu
    """
    deprec_warn("sFLH2sENU", "sigma_geo2enu")
    return sigma_geo2enu(*args, **kwargs)

def sFLH2sXYZ(*args, **kwargs):
    """
    Deprecated alias for sigma_geo2xyz
    """
    deprec_warn("sFLH2sXYZ", "sigma_geo2xyz")
    return sigma_geo2xyz(*args, **kwargs)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: mansur
This sub-module of geodezyx.operational contains functions to download
gnss data and products from distant IGS servers. 
it can be imported directly with:
from geodezyx import operational
The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License
Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

#### Import the logger
import logging
########## BEGIN IMPORT ##########
#### External modules
import os
import pandas as pd

#### geodeZYX modules
from geodezyx import conv
from geodezyx import utils

log = logging.getLogger('geodezyx')

##########  END IMPORT  ##########


## GUS CDDIS FCT FOR REPRO3 ONLY!!!! 2022-10-10
def list_file_in_ftp(
    email="mansur@gfz-potsdam.de",
    host="gdc.cddis.eosdis.nasa.gov",
    directory="gnss/products/repro3/1900",
):
    """
    List files inside one directory in a ftp serve. Note that this uses an anonymous user.

    Parameters
    ----------
    email : login email
    host  : ftp server
    directory  : directory inside the server
    Returns
    -------
    Return a list with the the information about the directory.

    """
    from ftplib import FTP_TLS

    # import sys
    # import ftplib
    ftps = FTP_TLS(host=host)
    ftps.login(user="anonymous", passwd=email)
    ftps.prot_p()
    ftps.cwd(directory)
    dir_list = []
    # print(ftps.dir(dir_list.append))
    ftps.dir(dir_list.append)

    return dir_list


def list_daily_files_ftp(ftp_dir_list=None, fmt="sp3", ac="COD", week=1900, dayw=0):
    """
    List files inside one directory in a ftp serve. Note that this uses an anonymous user.

    Parameters
    ----------
    ftp_dir_list : list with the files in the ftp (comming from the previous fct: list_file_in_ftp)
    fmt : searching format file e.g. sp3
    ac : name of the anlisys center
    week : gps week
    day : day of the week
    Returns
    -------
    Return a list with the the information about the directory.

    """

    ftp_dir_list = ftp_dir_list

    ## List all the files important e.g. SP3, BIA, SNX etc
    files_list = []
    for item in ftp_dir_list:
        splt_item = item.split()
        if splt_item[-1][-2:] not in ["gz"]:
            continue
        else:
            files_list.append(splt_item[-1])
    ##

    ## get all the format needed
    ## sp3
    if fmt in ["SP3", "EPH", "sp3", "eph"]:
        fmt = ["SP3", "EPH", "sp3", "eph"]
    ## clk
    if fmt in ["CLK", "clk"]:
        fmt = ["CLK", "clk"]
    fmt_list = []
    for item in files_list:
        if item[-6:-3] in fmt:
            fmt_list.append(item)

    ##

    ## get the desired AC
    ac = ac
    ac_list = []
    for item in fmt_list:
        if item[0:3] == ac:
            ac_list.append(item)
    ##

    ## Find the daily file
    week = week
    dayw = dayw
    day, year = conv.dt2doy_year(conv.gpstime2dt(week, dayw))
    string_date = str(year) + str(day)
    day_file_lst = []
    counter = 0
    for item in ac_list:
        if string_date in item:
            day_file_lst.append(item)
            counter = counter + 1
    if not day_file_lst:
        return None
    else:
        if counter > 1:
            print("WRN: More than one file found for this AC for this day!!")
            return day_file_lst
        else:
            return day_file_lst[0]
    ##


def download_rp3_cddis(week_ini=None, ac=None, fmt=None, out_path=None):
    """
    Download sp3 weekly files for repro3 products, from cddis (NASA).
    Note: For the moment is only for sp3 and clk.

    Parameters
    ----------
    week_ini : GPS week for download
    fmt : searching format file e.g. sp3
    ac : name of the anlisys center
    out_path : path where the products will be download

    Returns
    -------
    No returns.

    """
    ## arguments from the calling fct
    week_ini = week_ini
    week_final = week_ini + 1
    ac = ac
    fmt = fmt
    ##
    ## Default configuration using the user and cddis from NASA, on this case to repro3
    email = "mansur@gfz-potsdam.de"
    host = "gdc.cddis.eosdis.nasa.gov"
    directory = "gnss/products/repro3/"
    ##
    ## Set the output path. Not that os.chdir do a cd to the path. So
    ## the products are download to the right location
    out_path = out_path
    results_dir = os.path.join(out_path, "Week_" + str(week_ini))
    utils.create_dir(results_dir)
    os.chdir(results_dir)
    ##
    ## creates a list with the files for the week
    dir_wk = directory + str(week_ini)
    ftp_dir_list = list_file_in_ftp(email=email, host=host, directory=dir_wk)
    ##

    ## Do a loop for each day in the week
    for week in range(week_ini, week_final):
        for day in range(0, 7):
            try:
                file = list_daily_files_ftp(
                    ftp_dir_list=ftp_dir_list, fmt=fmt, ac=ac, week=week, dayw=day
                )
                if not file:
                    print("WRN: no file found!")
                    continue
                ## check if file is a list. This is the case when the AC has
                ## more than one solution for the day!! e.g. MIT
                if isinstance(file, list):
                    for fl in file:
                        cmd = (
                            "wget -qr -nH --cut-dirs=9 -t 10 --ftp-user anonymous --ftp-password "
                            + email
                            + " ftps://"
                            + host
                            + "/"
                            + dir_wk
                            + "/"
                            + fl
                        )
                        print(cmd)
                        os.system(cmd)
                else:
                    cmd = (
                        "wget -qr -nH --cut-dirs=9 -t 10 --ftp-user anonymous --ftp-password "
                        + email
                        + " ftps://"
                        + host
                        + "/"
                        + dir_wk
                        + "/"
                        + file
                    )
                    print(cmd)
                    os.system(cmd)

            except Exception as e:
                print(
                    "ERR: THERE IS AN ERROR WITH THIS DATE -- WEEK: "
                    + str(week)
                    + " DAY "
                    + str(day)
                )
                print(e)
                elm_stk = [week, day, e]
                err = pd.DataFrame(elm_stk)
                err = err.T
                err = err.rename(columns={0: "week", 1: "day", 2: "error"})
                pickle_name = "ERROR_" + str(week_ini) + "_" + str(day)
                err.to_pickle(results_dir + "/" + pickle_name + ".pkl")
                continue

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug  5 21:08:05 2023

@author: psakic

This module regroups the functions for the exploitation of the A0A pressure
sensors in the context of the REVOSIMA network

It is based on the work of Yann-Terden Tranchant, LIENSs, La Rochelle, France
"""


import numpy as np
import pandas as pd
import xarray as xr
# import gsw
#from scipy.signal import butter, filtfilt
#import scipy.signal as signal
#import scipy.optimize as optimize
import scipy


import datetime as dt


# from datetime import datetime, timedelta

#import numpy as np
#from scipy.signal import butter

GRAVITY = 9.81


def read_rbr_txt_data(file, t0 = None, t1 = None):

    col = ['BPR temperature', 'BPR pressure', 'BPR temperature.1',
           'BPR pressure.1', 'Temperature', 'Barometer temperature',
           'Barometer pressure']

    new_col = ['BPR temperature 1', 'BPR pressure 1', 'BPR temperature 2',
           'BPR pressure 2', 'Temperature', 'Barometer temperature',
           'Barometer pressure']

    A0A = pd.read_csv(file, index_col = 0, parse_dates = True)
    A0A = A0A[col]
    A0A.columns = new_col
    return A0A[t0:t1].dropna()


def butter_highpass(cutoff, fs, order=5):
    """
    Designs a highpass Butterworth filter.

    Parameters
    ----------
    cutoff : float
        The cutoff frequency of the filter.
    fs : float
        The sampling frequency of the signal.
    order : int, optional
        The order of the filter. Default is 5.

    Returns
    -------
    b : ndarray
        The numerator coefficient vector of the filter.
    a : ndarray
        The denominator coefficient vector of the filter.
    """
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    out_butter = scipy.signal.butter(order, normal_cutoff, btype='high', analog=False, output='ba')
    b, a = out_butter[0], out_butter[1]
    return b, a

def butter_lowpass(cutoff, fs, order=5):
    """
    Designs a lowpass Butterworth filter.

    Parameters
    ----------
    cutoff : float
        The cutoff frequency of the filter.
    fs : float
        The sampling frequency of the signal.
    order : int, optional
        The order of the filter. Default is 5.

    Returns
    -------
    b : ndarray
        The numerator coefficient vector of the filter.
    a : ndarray
        The denominator coefficient vector of the filter.
    """
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    out_butter = scipy.signal.butter(order, normal_cutoff, btype='low', analog=False, output='ba')
    b, a = out_butter[0], out_butter[1]

    return b, a

def butter_highpass_filtfilt(data, cutoff, fs, order=4):
    """
    Applies a highpass Butterworth filter to the data using forward and backward filtering.

    Parameters
    ----------
    data : array_like
        The input data to be filtered.
    cutoff : float
        The cutoff frequency of the filter.
    fs : float
        The sampling frequency of the signal.
    order : int, optional
        The order of the filter. Default is 4.

    Returns
    -------
    y : ndarray
        The filtered data.
    """
    b, a = butter_highpass(cutoff, fs, order=order)
    y = scipy.signal.filtfilt(b, a, data)
    return y

def butter_lowpass_filtfilt(data, cutoff, fs, order=4):
    """
    Applies a lowpass Butterworth filter to the data using forward and backward filtering.

    Parameters
    ----------
    data : array_like
        The input data to be filtered.
    cutoff : float
        The cutoff frequency of the filter.
    fs : float
        The sampling frequency of the signal.
    order : int, optional
        The order of the filter. Default is 4.

    Returns
    -------
    y : ndarray
        The filtered data.
    """
    b, a = butter_lowpass(cutoff, fs, order=order)
    y = scipy.signal.filtfilt(b, a, data)
    return y

def butter_bandpass(lowcut, highcut, fs, order=3):
    """
    Designs a bandpass Butterworth filter.

    Parameters
    ----------
    lowcut : float
        The lower cutoff frequency of the filter.
    highcut : float
        The upper cutoff frequency of the filter.
    fs : float
        The sampling frequency of the signal.
    order : int, optional
        The order of the filter. Default is 3.

    Returns
    -------
    b : ndarray
        The numerator coefficient vector of the filter.
    a : ndarray
        The denominator coefficient vector of the filter.
    """
    return scipy.signal.butter(order, [lowcut, highcut], fs=fs, btype='band')

def butter_bandpass_filtfilt(data, lowcut, highcut, fs, order=4):
    """
    Applies a bandpass Butterworth filter to the data using forward and backward filtering.

    Parameters
    ----------
    data : array_like
        The input data to be filtered.
    lowcut : float
        The lower cutoff frequency of the filter.
    highcut : float
        The upper cutoff frequency of the filter.
    fs : float
        The sampling frequency of the signal.
    order : int, optional
        The order of the filter. Default is 4.

    Returns
    -------
    y : ndarray
        The filtered data.
    """
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = scipy.signal.filtfilt(b, a, data)
    return y

def butterworth(df, t0=3*24*3600, t1=10*24*3600, kind='bandpass', order=4):
    """
    Applies a Butterworth filter to the data.

    Parameters
    ----------
    df : pandas.DataFrame
        The input data to be filtered.
    t0 : int, optional
        The lower cutoff period in seconds. Default is 3 days (3*24*3600 seconds).
    t1 : int, optional
        The upper cutoff period in seconds. Default is 10 days (10*24*3600 seconds).
    kind : str, optional
        The type of filter to apply. Options are 'highpass', 'lowpass', and 'bandpass'. Default is 'bandpass'.
    order : int, optional
        The order of the filter. Default is 4.

    Returns
    -------
    pandas.Series
        The filtered data.
    """
    difftime = np.diff(df.index)
    if not df.index[0].tzname():
        pass  # directly nanosec
    else:
        difftime = np.array([e.total_seconds() * 10**9 for e in difftime])  # TimeDelta to be conv in nanosec

    if not np.all(difftime == difftime[0]):
        print('The sampling is not regular, return')
        return
    else:
        fs = 1/(float(difftime[0])/1e9)

    if kind == 'highpass':
        cut = 1/t0
        y = butter_highpass_filtfilt(df.values, cut, fs, order=order)
    elif kind == 'lowpass':
        cut = 1/t0
        y = butter_lowpass_filtfilt(df.values, cut, fs, order=order)
    elif kind == 'bandpass':
        lowcut = 1/t1
        highcut = 1/t0
        y = butter_bandpass_filtfilt(df.values, lowcut, highcut, fs, order=order)
    else:
        print(f'Invalid filter type : {kind}. Must be highpass, lowpass, or bandpass.')
        return

    return pd.DataFrame(y, index=df.index)[0]

def read_hycom(file):
    """
    Reads a HYCOM dataset and renames its variables.

    Parameters
    ----------
    file : str
        The path to the HYCOM dataset file.

    Returns
    -------
    xarray.Dataset
        The renamed dataset.
    """
    ds = xr.open_dataset(file)
    return ds.rename(lon='longitude', lat='latitude', water_temp='theta', salinity='salt', surf_el='ssh')

def read_ecco2(file):
    """
    Reads an ECCO2 dataset and renames its variables.

    Parameters
    ----------
    file : str
        The path to the ECCO2 dataset file.

    Returns
    -------
    xarray.Dataset
        The renamed dataset.
    """
    ds = xr.open_dataset(file)
    return ds.rename(LEV='depth')

def read_glorys(file):
    """
    Reads a GLORYS dataset and renames its variables.

    Parameters
    ----------
    file : str
        The path to the GLORYS dataset file.

    Returns
    -------
    xarray.Dataset
        The renamed dataset.
    """
    ds = xr.open_dataset(file, decode_times=True)
    return ds.rename(thetao='theta', so='salt', zos='ssh')

def read_duacs(file):
    """
    Reads a DUACS dataset and renames its variables.

    Parameters
    ----------
    file : str
        The path to the DUACS dataset file.

    Returns
    -------
    xarray.Dataset
        The renamed dataset.
    """
    ds = xr.open_dataset(file)
    return ds.rename(sla='ssh')

def interp_xy(ds, x=None, y=None, method = 'linear'):
    """
    frontend for xarray spatial interpolation

    Parameters
    ----------
    ds : xarray.Dataset
        input xarray grid.
    x : float, optional
        longitude in deg (but depends on ds units). The default is None.
    y : float, optional
        latitude in deg (but depends on ds units). The default is None.
    method : str or Litteral, optional
        interpolation method. The default is 'linear'.

    Returns
    -------
    ds
        Interpolated values.

    """
    if not x and not y:
        print('Must give x or/and y coord')
        return
    if not x:
        return ds.interp(latitude = y, method = method)
    elif not y:
        return ds.interp(longitude = x, method = method)
    else:
        return ds.interp(longitude = x, latitude = y, method = method)


def interp_time(ds, time, method='linear'):
    """
    Interpolates the dataset along the time dimension.

    Parameters
    ----------
    ds : xarray.Dataset
        The input dataset to be interpolated.
    time : array-like or scalar
        The new time points to interpolate to.
    method : str or Litteral, optional
        The interpolation method to use. Default is 'linear'.

    Returns
    -------
    xarray.Dataset
        The interpolated dataset.
    """
    return ds.interp(time=time, method=method)

def interp_z(ds, depth, method='linear'):
    """
    Interpolates the dataset along the depth dimension.

    Parameters
    ----------
    ds : xarray.Dataset
        The input dataset to be interpolated.
    depth : array-like or scalar
        The new depth points to interpolate to.
    method : str or Litteral, optional
        The interpolation method to use. Default is 'linear'.

    Returns
    -------
    xarray.Dataset
        The interpolated dataset.
    """
    return ds.interp(depth=depth, method=method)

def get_bottom_depth(ds):
    """
    Computes the bottom depth of the dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        The input dataset containing depth and theta variables.

    Returns
    -------
    numpy.ndarray
        The bottom depth values.
    """
    land = np.isnan(ds.theta[0, 0, :, :])
    isnan = np.isnan(ds.theta[0, 0:, :, :])
    ind_nan = np.count_nonzero(~isnan, axis=0)
    bottom_depth = ds.depth.values[ind_nan]
    bottom_depth[land.values] = np.nan
    return bottom_depth

def extract_profile(ds, x, y, method='linear', zbottom=None):
    """
    Extracts a profile from the dataset at the specified coordinates and interpolates to the given depth.

    Parameters
    ----------
    ds : xarray.Dataset
        The input dataset containing the data.
    x : float
        The longitude coordinate.
    y : float
        The latitude coordinate.
    method : str, optional
        The interpolation method to use. Default is 'linear'.
    zbottom : int, float, list, or numpy.ndarray, optional
        The depth(s) to interpolate to. If None, no depth interpolation is performed. Default is None.

    Returns
    -------
    xarray.Dataset
        The extracted and interpolated profile.
    """
    ds = interp_xy(ds, x, y, method)

    if not isinstance(zbottom, type(None)):
        if isinstance(zbottom, (int, float)):
            valid = ~np.isnan(ds.theta[0])
            max_depth = ds.depth[valid][-1]
            if zbottom > max_depth.values:
                print(f'{zbottom} m is deeper than the maximum depth layer of the profile : {max_depth.values}')
                return None
            k = len((zbottom - ds.depth)[(zbottom - ds.depth) > 0]) - 1
            new_depth = np.hstack([ds.depth[:k], zbottom])
        elif isinstance(zbottom, (np.ndarray, list)):
            new_depth = zbottom
        else:
            print(f'Invalid zbottom parameter type : {type(zbottom)}. Must be int or float depth (in m).')
            return None
        ds = interp_z(ds, new_depth, method='linear')

    return ds

def compute_dens_profile(profile):
    """
    Computes the density profile from the given dataset.

    Parameters
    ----------
    profile : xarray.Dataset
        The input dataset containing temperature (theta) and salinity (salt) variables.

    Returns
    -------
    numpy.ndarray
        The computed density profile.
    """
    import seawater
    ds = profile
    if np.ndim(ds.theta) != 2:
        print(f'Bad dataset dimension : {np.ndim(ds.theta)}. The function must be applied on a profile.')
        return None
    lat = ds.latitude.mean()
    pres = seawater.pres(ds.depth, lat=lat)
    dens = seawater.dens(ds.salt, ds.theta, pres)
    return dens

def compute_steric(profile, integration='forward'):
    """
    Computes the steric height anomaly from the given profile.

    Parameters
    ----------
    profile : xarray.Dataset
        The input dataset containing temperature (theta) and salinity (salt) variables.
    integration : str, optional
        The integration method to use. Options are 'forward', 'backward', and 'averaged'. Default is 'forward'.

    Returns
    -------
    pandas.Series
        The steric height anomaly.
    """
    ds = profile
    #lat = ds.latitude.mean()
    dens = compute_dens_profile(ds)
    if integration == 'forward':
        steric = pd.DataFrame(GRAVITY * np.nansum(np.diff(ds.depth) * dens[:, :-1], axis=1) / 10000, index=ds.time)[0]
    elif integration == 'backward':
        steric = pd.DataFrame(GRAVITY * np.nansum(np.diff(ds.depth) * dens[:, 1:], axis=1) / 10000, index=ds.time)[0]
    elif integration == 'averaged':
        dens = np.vstack([np.sum(dens[:, i:i+2], axis=1) / 2 for i in range(len(ds.depth) - 1)]).T
        steric = pd.DataFrame(GRAVITY * np.nansum(np.diff(ds.depth) * dens, axis=1) / 10000, index=ds.time)[0]
    else:
        print(f'Invalid integration method : {integration}. Must be forward, backward, or averaged.')
        return None

    return steric - steric.median()

def compute_phibot(profile, ssh=None, integration='forward', rho=10.35, remove_median=True):
    """
    Computes the ocean hydrostatic bottom pressure anomaly (phibot).

    Parameters
    ----------
    profile : xarray.Dataset
        The input dataset containing temperature (theta) and salinity (salt) variables.
    ssh : xarray.DataArray, optional
        The sea surface height data. If None, it is taken from the profile. Default is None.
    integration : str, optional
        The integration method to use for steric height computation.
        Options are 'forward', 'backward', and 'averaged'. Default is 'forward'.
    rho : float, optional
        The reference density in kg/m^3. Default is 10.35.
    remove_median : bool, optional
        Whether to remove the median from the computed phibot values. Default is True.

    Returns
    -------
    pandas.DataFrame
        The computed phibot values.

    Notes
    -----
    phibot = Ocean hydrostatic bottom pressure anomaly
    https://cmr.earthdata.nasa.gov/search/concepts/V2028471168-POCLOUD.html
    https://cmr.earthdata.nasa.gov/search/concepts/V2146301108-POCLOUD.html


    PHIBOT		Bottom Pressure Pot. Anomaly (p/rhonil, m^2/s^2)
                To convert to m, divide by g (g=9.81 m/s^2)
		PHIBOT is the anomaly relative to Depth * rhonil * g
		The absolute bottom pressure in Pa is:
		Depth * rhonil * g + PHIBOT * rhonil (rhonil=1027.5 kg/m^3)
    http://apdrc.soest.hawaii.edu/doc/Readme_ecco2_cube92
    """
    if isinstance(ssh, type(None)):
        ssh = profile.ssh
    #lat = profile.latitude.mean()
    #dens = compute_dens_profile(profile)
    phibot = pd.DataFrame(index=profile.time)
    phibot['subsurface'] = compute_steric(profile, integration)
    phibot['surface'] = GRAVITY * rho * ssh / 100
    phibot['phibot'] = phibot.subsurface + phibot.surface
    if remove_median:
        phibot -= phibot.median(axis=0)

    return phibot
    
def compute_spectrogram(df, max_period=45, nchunks=3600*6):
    """
    Computes the spectrogram of the given data.

    Parameters
    ----------
    df : pandas.DataFrame
        The input data to compute the spectrogram for.
    max_period : int, optional
        The maximum period to consider in the spectrogram. Default is 45.
    nchunks : int, optional
        The number of chunks to divide the data into. Default is 3600*6.

    Returns
    -------
    p : numpy.ndarray
        The periods corresponding to the spectrogram.
    t : pandas.DatetimeIndex
        The time points corresponding to the spectrogram.
    sxx : numpy.ndarray
        The spectrogram values.
    """
    from scipy.signal import spectrogram

    f, t, sxx = spectrogram(df.values, nperseg=nchunks, scaling='density')
    p = (1/f[1:])
    sxx = sxx[1:]
    i = np.argmin(np.abs(p - max_period))
    p = p[i:]
    sxx = sxx[i:]

    t = pd.DatetimeIndex([df.index[0] + dt.timedelta(hours=t) for t in t/3600])
    return p, t, sxx

def log_linear(x, a1, a2, a3, a4, **kwargs):
    """
    Log-linear model function.

    Parameters
    ----------
    x : array-like
        The input data.
    a1 : float
        Coefficient for the logarithmic term.
    a2 : float
        Coefficient for the logarithmic term.
    a3 : float
        Coefficient for the linear term.
    a4 : float
        Constant term.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    float
        The computed log-linear value.
    """
    return a1 * np.log(a2 * (1 + x)) + a3 * x + a4

def log(x, a1, a2, a3, a4, **kwargs):
    """
    Logarithmic model function.

    Parameters
    ----------
    x : array-like
        The input data.
    a1 : float
        Coefficient for the logarithmic term.
    a2 : float
        Coefficient for the logarithmic term.
    a3 : float
        Constant term.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    float
        The computed logarithmic value.
    """
    x0 = min(x)
    return a1 * np.log(a2 * (1 + x - x0)) + a3 * (x - x0) + a4

def exp(x, a, b, c, d, **kwargs):
    """
    Exponential model function.

    Parameters
    ----------
    x : array-like
        The input data.
    a : float
        Coefficient for the exponential term.
    b : float
        Coefficient for the exponential term.
    c : float
        Constant term.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    float
        The computed exponential value.
    """
    return a * np.exp(-b * x) + c * x + d 

def linear(x, a1, a2, a3, a4, **kwargs):
    """
    Linear model function.

    Parameters
    ----------
    x : array-like
        The input data.
    a1 : float
        Slope of the linear function.
    a2 : float
        Intercept of the linear function.
    a3 : float
        Additional coefficient.
    a4 : float
        Additional coefficient.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    float
        The computed linear value.
    """
    x0 = min(x)
    return a1 * (x - x0) + a2

def exp_linear(x, a, b, c, d, **kwargs):
    """
    Exponential-linear model function.

    Parameters
    ----------
    x : array-like
        The input data.
    a : float
        Coefficient for the exponential term.
    b : float
        Coefficient for the exponential term.
    c : float
        Linear term coefficient.
    d : float
        Constant term.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    float
        The computed exponential-linear value.
    """
    x0 = min(x)
    return a * np.exp(-1/b * (x - x0)) + c * (x - x0) + d


#### Added by PS (2024-10)
def exp_polynomial_deg2(x, a, b, c,  **kwargs):
    """
    Computes an exponential function combined with a second-degree polynomial.

    This function calculates the value of an exponential function combined with a
    second-degree polynomial given the input data and coefficients.

    Parameters
    ----------
    x : array-like
        The input data.
    a : float
        Coefficient for the polynomial term.
    b : float
        Coefficient for the exponential term.
    c : float
        Constant term.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    float
        The computed value from the exponential function combined with a second-degree polynomial.
    """
    x0 = min(x)
    return (x - x0)**2 * a * np.exp(b * (x-x0) + c)

# Exponential with offset
def exp_with_offset(x, a, b, c, **kwargs):
    """
    Exponential model function with an offset.

    This function calculates an exponential decay with an added constant offset.

    Parameters
    ----------
    x : array-like
        The input data.
    a : float
        Coefficient for the exponential term.
    b : float
        Exponent for the exponential term.
    c : float
        Constant offset term.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    float
        The computed value from the exponential model with offset.
    """
    return a * np.exp(-b * x) + c

# Higher degree polynomial (3rd degree)
def polynomial_3(x, a, b, c, d, **kwargs):
    """
    Computes a 3rd degree polynomial.

    This function calculates the value of a 3rd degree polynomial given the input data and coefficients.

    Parameters
    ----------
    x : array-like
        The input data.
    a : float
        The constant term of the polynomial.
    b : float
        The coefficient of the linear term.
    c : float
        The coefficient of the quadratic term.
    d : float
        The coefficient of the cubic term.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    float
        The computed value of the 3rd degree polynomial.
    """
    return a + b*x + c*x**2 + d*x**3
# Generalized nonlinear model (exponential + power law)
def exp_power_combined(x, a, b, c, d, **kwargs):
    """
    Generalized nonlinear model function combining exponential decay and power law.

    This function models data using a combination of an exponential decay term and a power law term.

    Parameters
    ----------
    x : array-like
        The input data.
    a : float
        Coefficient for the exponential term.
    b : float
        Exponent for the exponential term.
    c : float
        Coefficient for the power law term.
    d : float
        Exponent for the power law term.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    float
        The computed value from the combined exponential and power law model.
    """
    return a * np.exp(-b * x) + c * np.power(x + 1, -d)

def fit_model(data, t_fit, model='log_linear', offset_last=True, maxfev=1000, pn=2,
              normalize_t=False):
    """
    Fits a model to the given data.

    Parameters
    ----------
    data : pandas.Series
        The input data to fit the model to.
        Is a Series with time as Index
    t_fit : pandas.DatetimeIndex
        Time points to get fitted values once the fit model has been determeined.
    model : str, optional
        The model to use for fitting. Options are 'log_linear', 'log', 'exp',
        'linear', 'exp_linear', and 'poly'. Default is 'log_linear'.
    offset_last : bool, optional
        Whether to offset the last value. Default is True.
    maxfev : int, optional
        The maximum number of function evaluations. Default is 1000.
    pn : int, optional
        The degree of the polynomial if the model is 'poly'. Default is 2.
    normalize_t : bool, optional
        Whether to normalize the time values. Default is False.

    Returns
    -------
    pandas.Series
        The fitted model values.
    """

    # initial time implemented in days, but better in seconds
    #t_num_data = (data.index.to_julian_date() - data.index.to_julian_date()[0]).values
    #t_num_fit = (t_fit.to_julian_date() - t_fit.to_julian_date()[0]).values

    t_num_data = (data.index - data.index[0]).total_seconds().values
    t_num_fit = (t_fit - t_fit[0]).total_seconds().values

    ### normalize
    if normalize_t:
        ref = t_num_data[-1]
        t_num_data /= ref
        t_num_fit /= ref

    if model == 'poly':
        z = np.polyfit(t_num_data, data.values, pn)
        p = np.poly1d(z)
        fit = p(t_num_fit)
        ind_min = np.argmin(fit)
        fit[ind_min:] = fit[ind_min]
        out_opti = None
    elif model == 'log_linear':
        out_opti = optimize.curve_fit(log_linear, t_num_data, data.values, maxfev=maxfev)
        popt, pcov = out_opti[0], out_opti[1]
        fit = log_linear(t_num_fit, *popt)
    elif model == 'log':
        out_opti = optimize.curve_fit(log, t_num_data, data.values, maxfev=maxfev)
        popt, pcov = out_opti[0], out_opti[1]
        fit = log(t_num_fit, *popt)
    elif model == 'exp':
        out_opti = optimize.curve_fit(exp, t_num_data, data.values, maxfev=maxfev)
        popt, pcov = out_opti[0], out_opti[1]
        fit = exp(t_num_fit, *popt)
    elif model == 'exp_linear':
        out_opti = optimize.curve_fit(exp_linear, t_num_data, data.values, maxfev=maxfev)
        popt, pcov = out_opti[0], out_opti[1]
        fit = exp_linear(t_num_fit, *popt)
    elif model == 'exp_polynomial_deg2':
        out_opti = optimize.curve_fit(exp_polynomial_deg2, t_num_data, data.values, maxfev=maxfev)
        popt, pcov = out_opti[0], out_opti[1]
        fit = exp_polynomial_deg2(t_num_fit, *popt)
    elif model == 'linear':
        out_opti = optimize.curve_fit(linear, t_num_data, data.values, maxfev=maxfev)
        popt, pcov = out_opti[0], out_opti[1]
        fit = linear(t_num_fit, *popt)
    elif model == 'exp_power_combined':
        out_opti = optimize.curve_fit(exp_power_combined, t_num_data, data.values, maxfev=maxfev)
        popt, pcov = out_opti[0], out_opti[1]
        fit = exp_power_combined(t_num_fit, *popt)
    else:
        print(f'Invalid model type : {model}. Must be log_linear, log, exp, linear, exp_linear, or poly.')
        return None, None

    return pd.DataFrame(fit, index=t_fit)[0], out_opti


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# https://chatgpt.com/c/683ec5ef-d6d0-8012-8163-dc4e8a6297b3
"""
Created on Tue Jun  3 11:52:28 2025

@author: psakicki
"""
from netCDF4 import Dataset
import numpy as np
from datetime import datetime, timedelta, UTC

# Create NetCDF file
ncfile = Dataset('insitu_cmems_example.nc', mode='w', format='NETCDF4')

# Define global attributes (partial list per CMEMS/SeaDataNet convention)
ncfile.title = "CMEMS In Situ TAC Example"
ncfile.institution = "Your Institution Name"
ncfile.source = "In situ observations"
ncfile.history = f"Created on {datetime.now(UTC).isoformat()} UTC"
ncfile.references = "http://marine.copernicus.eu"
ncfile.comment = "Prototype NetCDF file for CMEMS In Situ TAC"
ncfile.data_type = "OceanSITES time-series data"
ncfile.Conventions = "CF-1.6, OceanSITES-1.3, SeaDataNet"

# Define dimensions
time_dim = ncfile.createDimension('TIME', None)  # Unlimited
depth_dim = ncfile.createDimension('DEPTH', 1)
lat_dim = ncfile.createDimension('LATITUDE', 1)
lon_dim = ncfile.createDimension('LONGITUDE', 1)

# Define coordinate variables
times = ncfile.createVariable('TIME', 'f8', ('TIME',))
times.units = 'days since 1950-01-01 00:00:00 UTC'
times.standard_name = 'time'
times.long_name = 'Time of measurement'
times.axis = 'T'

depth = ncfile.createVariable('DEPTH', 'f4', ('DEPTH',))
depth.units = 'm'
depth.positive = 'down'
depth.standard_name = 'depth'
depth.long_name = 'Depth of measurement'
depth.axis = 'Z'

lat = ncfile.createVariable('LATITUDE', 'f8', ('LATITUDE',))
lat.units = 'degrees_north'
lat.standard_name = 'latitude'
lat.long_name = 'Latitude'
lat.axis = 'Y'

lon = ncfile.createVariable('LONGITUDE', 'f8', ('LONGITUDE',))
lon.units = 'degrees_east'
lon.standard_name = 'longitude'
lon.long_name = 'Longitude'
lon.axis = 'X'

# Define a sample data variable (e.g., temperature)
temp = ncfile.createVariable('TEMP', 'f4', 
                             ('TIME', 'DEPTH', 'LATITUDE', 'LONGITUDE',),
                             fill_value=-9999.0)

temp.units = 'degree_Celsius'
temp.standard_name = 'sea_water_temperature'
temp.long_name = 'Sea Water Temperature'
temp.coordinates = "TIME DEPTH LATITUDE LONGITUDE"

# Fill data
n_points = 10
base_time = datetime(2023, 1, 1)
times[:] = [(base_time + timedelta(days=i) - datetime(1950, 1, 1)).days for i in range(n_points)]
depth[:] = [10.0]
lat[:] = [42.0]
lon[:] = [3.0]
temp[:, 0, 0, 0] = np.linspace(14.5, 15.5, n_points)  # Mock temperature data

# Close the file
ncfile.close()
print("CMEMS In Situ TAC NetCDF prototype created.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 21/03/2025 21:57:04

@author: psakic
"""

import geodezyx.operational as opera
import geodezyx.conv as conv
import datetime as dt
import argparse
import os


def singugins_run(
    results_folder,
    specific_sites=[],
    start_epoch=dt.datetime(2000, 5, 3),
    end_epoch=dt.datetime(2099, 12, 31),
    bdrnx_folder="/root/020_BDRNX/",
    nprocs=8,
    no_updatebd=False,
    no_concat_orb_clk=False,
    verbose=False,
    force=False,
    spotgins_run_kwargs={},
    no_rnx2=False,
    no_rnx3=False,
):
    """
    Run the SPOTGINS process within a SINGUGINS container.

    Parameters
    ----------
    results_folder : str
        Path to the folder where the results will be stored.
        We strongly recommend using a subfolder within
        the default SINGUGINS results folder
        `/root/030_RESULTS/<your_subfolder>`.
    specific_sites : list, optional
        List of specific RINEXs site IDs to include in the process.
        Default is an empty list.
    start_epoch : datetime, optional
        Start date and time for the selected RINEXs. Default is May 3, 2000.
    end_epoch : datetime, optional
        End date and time for the selected RINEXs. Default is December 31, 2099.
    bdrnx_folder : str, optional
        Path to the folder containing the RINEX files. Default is "/root/020_BDRNX/".
    nprocs : int, optional
        Number of processes to use. Default is 8.
    no_updatebd : bool, optional
        Flag to indicate whether to update the database. Default is False.
    no_concat_orb_clk : bool, optional
        Flag to indicate whether to concatenate the orbit and
        clock files prior to the main processing
        Default is False.
    verbose : bool, optional
        Verbose mode. If True, additional information will be printed.
    force : bool, optional
        Flag to force the process to run even if the results already exist.
    spotgins_run_kwargs : dict, optional
        Additional keyword arguments to pass to the `spotgins_run` function.
        Default is an empty dictionary.
    no_rnx2 : bool, optional
        If True, RINEX2 files will not be processed.
        Default is False (RINEX2 files will be processed).
    no_rnx3 : bool, optional
        If True, RINEX3 files will not be processed.
        Default is False (RINEX3 files will be processed).

    Returns
    -------
    None
    """

    import numpy as np
    srt_epoch_ok = min([start_epoch, end_epoch])
    end_epoch_ok = max([start_epoch, end_epoch])

    rnxs_lis = opera.rinex_finder(
        bdrnx_folder,
        specific_sites=specific_sites,
        start_epoch=srt_epoch_ok,
        end_epoch=end_epoch_ok,
        short_name=not no_rnx2,
        long_name=not no_rnx3,
    )

    opera.spotgins_run(
        rnxs_path_inp=rnxs_lis,
        results_folder_inp=results_folder,
        nprocs=nprocs,
        no_updatebd=no_updatebd,
        no_concat_orb_clk=no_concat_orb_clk,
        verbose=verbose,
        force=force,
        **spotgins_run_kwargs,
    )

    return

def main():
    parser = argparse.ArgumentParser(
        description="Run the SPOTGINS process within a SINGUGINS container."
    )
    parser.add_argument(
        "-o",
        "--results_folder",
        type=str,
        help="Path to the folder where the results will be stored.",
        required=True,
    )
    parser.add_argument(
        "-l",
        "--specific_sites",
        type=str,
        nargs="*",
        default=[],
        help="List of specific RINEXs site IDs to include in the process.",
    )
    parser.add_argument(
        "-s",
        "--start_epoch",
        type=lambda s: conv.date_pattern_2_dt(s),
        default=dt.datetime(2000, 5, 3),
        help="Start date and time for the selected RINEXs.",
    )
    parser.add_argument(
        "-e",
        "--end_epoch",
        type=lambda s: conv.date_pattern_2_dt(s),
        default=dt.datetime(2099, 12, 31),
        help="End date and time for the selected RINEXs.",
    )
    parser.add_argument(
        "-r",
        "--bdrnx_folder",
        type=str,
        default="/root/020_BDRNX/",
        help="Path to the folder containing the RINEX files.",
    )
    parser.add_argument(
        "-k",
        "--spotgins_run_kwargs",
        type=str,
        default="{}",
        help="Additional keyword arguments to pass to the `spotgins_run` function (as a JSON string).",
    )

    parser.add_argument(
        "-p",
        "--nprocs",
        type=int,
        default=8,
        help="Number of processes to use.",
    )
    parser.add_argument(
        "-nu",
        "--no_updatebd",
        action="store_true",
        help="Flag to indicate whether to update the database.",
    )

    parser.add_argument(
        "-nc",
        "--no_concat_orb_clk",
        action="store_true",
        help="Flag to indicate whether to concatenate the"
             "orbit and clock files prior to the main processing",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Verbose mode.",
    )

    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Flag to force the process to run even if the results already exist.",
    )

    parser.add_argument(
        "-nr2",
        "--no_rnx2",
        action="store_true",
        help="If True, RINEX2 files will not be processed.",
    )

    parser.add_argument(
        "-nr3",
        "--no_rnx3",
        action="store_true",
        help="If True, RINEX3 files will not be processed.",
    )

    args = parser.parse_args()

    # Convert spotgins_run_kwargs from JSON string to dictionary
    import json

    spotgins_run_kwargs = json.loads(args.spotgins_run_kwargs)

    singugins_run(
        results_folder=os.path.abspath(args.results_folder),
        specific_sites=args.specific_sites,
        start_epoch=args.start_epoch,
        end_epoch=args.end_epoch,
        bdrnx_folder=args.bdrnx_folder,
        spotgins_run_kwargs=spotgins_run_kwargs,
        no_updatebd=args.no_updatebd,
        no_concat_orb_clk=args.no_concat_orb_clk,
        nprocs=args.nprocs,
        verbose=args.verbose,
        force=args.force,
        no_rnx2=args.no_rnx2,
        no_rnx3=args.no_rnx3,
    )


if __name__ == "__main__":
    main()
